from __future__ import annotations

import logging
from collections.abc import Callable, Iterable, Sequence
from typing import Any, cast

from lightning.pytorch import Callback, LightningModule
from typing_extensions import TypeAliasType, override

from ..._callback import NTCallbackBase
from ...util.typing_utils import mixin_base_type

log = logging.getLogger(__name__)

_Callback = TypeAliasType("_Callback", Callback | NTCallbackBase)
CallbackFn = TypeAliasType(
    "CallbackFn", Callable[[], _Callback | Iterable[_Callback] | None]
)


class CallbackRegistrarModuleMixin:
    @property
    def _nshtrainer_callbacks(self) -> list[CallbackFn]:
        if not hasattr(self, "_private_nshtrainer_callbacks_list"):
            self._private_nshtrainer_callbacks_list = []
        return self._private_nshtrainer_callbacks_list

    def register_callback(
        self,
        callback: _Callback | Iterable[_Callback] | CallbackFn | None = None,
    ):
        if not callable(callback):
            callback_ = cast(CallbackFn, lambda: callback)
        else:
            callback_ = callback

        self._nshtrainer_callbacks.append(callback_)


class CallbackModuleMixin(
    CallbackRegistrarModuleMixin,
    mixin_base_type(LightningModule),
):
    @property
    def _nshtrainer_callbacks(self) -> list[CallbackFn]:
        if not hasattr(self, "_private_nshtrainer_callbacks_list"):
            self._private_nshtrainer_callbacks_list = []
        return self._private_nshtrainer_callbacks_list

    def register_callback(
        self,
        callback: _Callback | Iterable[_Callback] | CallbackFn | None = None,
    ):
        if not callable(callback):
            callback_ = cast(CallbackFn, lambda: callback)
        else:
            callback_ = callback

        self._nshtrainer_callbacks.append(callback_)

    def _gather_all_callbacks(self):
        modules: list[Any] = []
        if isinstance(self, CallbackRegistrarModuleMixin):
            modules.append(self)
        if (
            datamodule := getattr(self.trainer, "datamodule", None)
        ) is not None and isinstance(datamodule, CallbackRegistrarModuleMixin):
            modules.append(datamodule)
        modules.extend(
            module
            for module in self.children()
            if isinstance(module, CallbackRegistrarModuleMixin)
        )
        for module in modules:
            yield from module._nshtrainer_callbacks

    @override
    def configure_callbacks(self):
        callbacks = super().configure_callbacks()
        if not isinstance(callbacks, Sequence):
            callbacks = [callbacks]

        callbacks = list(callbacks)
        for callback_fn in self._gather_all_callbacks():
            if (callback_result := callback_fn()) is None:
                continue

            if not isinstance(callback_result, Iterable):
                callback_result = [callback_result]

            for callback in callback_result:
                log.info(
                    f"Registering {callback.__class__.__qualname__} callback {callback}"
                )
                callbacks.append(callback)

        return callbacks



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer.plugin import PluginConfig as PluginConfig
from nshtrainer.trainer.plugin import PluginConfigBase as PluginConfigBase
from nshtrainer.trainer.plugin import plugin_registry as plugin_registry
from nshtrainer.trainer.plugin.environment import (
    KubeflowEnvironmentPlugin as KubeflowEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    LightningEnvironmentPlugin as LightningEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    LSFEnvironmentPlugin as LSFEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    MPIEnvironmentPlugin as MPIEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    SLURMEnvironmentPlugin as SLURMEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    TorchElasticEnvironmentPlugin as TorchElasticEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    XLAEnvironmentPlugin as XLAEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.io import (
    AsyncCheckpointIOPlugin as AsyncCheckpointIOPlugin,
)
from nshtrainer.trainer.plugin.io import (
    TorchCheckpointIOPlugin as TorchCheckpointIOPlugin,
)
from nshtrainer.trainer.plugin.io import XLACheckpointIOPlugin as XLACheckpointIOPlugin
from nshtrainer.trainer.plugin.layer_sync import (
    TorchSyncBatchNormPlugin as TorchSyncBatchNormPlugin,
)
from nshtrainer.trainer.plugin.precision import (
    BitsandbytesPluginConfig as BitsandbytesPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    DeepSpeedPluginConfig as DeepSpeedPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    DoublePrecisionPluginConfig as DoublePrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import DTypeConfig as DTypeConfig
from nshtrainer.trainer.plugin.precision import (
    FSDPPrecisionPluginConfig as FSDPPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    HalfPrecisionPluginConfig as HalfPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    MixedPrecisionPluginConfig as MixedPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    TransformerEnginePluginConfig as TransformerEnginePluginConfig,
)
from nshtrainer.trainer.plugin.precision import XLAPluginConfig as XLAPluginConfig

from . import base as base
from . import environment as environment
from . import io as io
from . import layer_sync as layer_sync
from . import precision as precision

__all__ = [
    "AsyncCheckpointIOPlugin",
    "BitsandbytesPluginConfig",
    "DTypeConfig",
    "DeepSpeedPluginConfig",
    "DoublePrecisionPluginConfig",
    "FSDPPrecisionPluginConfig",
    "HalfPrecisionPluginConfig",
    "KubeflowEnvironmentPlugin",
    "LSFEnvironmentPlugin",
    "LightningEnvironmentPlugin",
    "MPIEnvironmentPlugin",
    "MixedPrecisionPluginConfig",
    "PluginConfig",
    "PluginConfigBase",
    "SLURMEnvironmentPlugin",
    "TorchCheckpointIOPlugin",
    "TorchElasticEnvironmentPlugin",
    "TorchSyncBatchNormPlugin",
    "TransformerEnginePluginConfig",
    "XLACheckpointIOPlugin",
    "XLAEnvironmentPlugin",
    "XLAPluginConfig",
    "base",
    "environment",
    "io",
    "layer_sync",
    "plugin_registry",
    "precision",
]



from __future__ import annotations

import logging
from typing import Any

import torch

log = logging.getLogger(__name__)


def _trainer(module: Any):
    if torch.jit.is_scripting():
        return None

    if hasattr(module, "_trainer"):
        trainer = module._trainer
    else:
        try:
            trainer = module.trainer
        except RuntimeError:
            return None

    from ...trainer import Trainer

    if not isinstance(trainer, Trainer):
        return None

    return trainer


class DebugModuleMixin:
    @property
    def nshtrainer_or_none(self):
        return _trainer(self)

    @property
    def nshtrainer(self):
        if (trainer := _trainer(self)) is None:
            raise RuntimeError("Could not resolve trainer.")
        return trainer

    @property
    def debug(self) -> bool:
        if (trainer := _trainer(self)) is None:
            return False
        return trainer.debug

    @debug.setter
    def debug(self, value: bool):
        if (trainer := _trainer(self)) is None:
            return
        trainer.debug = value

    @torch.jit.unused
    def breakpoint(self, rank_zero_only: bool = True):
        if (
            not rank_zero_only
            or not torch.distributed.is_initialized()
            or torch.distributed.get_rank() == 0
        ):
            breakpoint()

        if rank_zero_only and torch.distributed.is_initialized():
            _ = torch.distributed.barrier()

    @torch.jit.unused
    def ensure_finite(
        self,
        tensor: torch.Tensor,
        name: str | None = None,
        throw: bool = False,
    ):
        name_parts: list[str] = ["Tensor"]
        if name is not None:
            name_parts.append(name)
        name = " ".join(name_parts)

        not_finite = ~torch.isfinite(tensor)
        if not_finite.any():
            msg = f"{name} has {not_finite.sum().item()}/{not_finite.numel()} non-finite values."
            if throw:
                raise RuntimeError(msg)
            else:
                log.warning(msg)
            return False
        return True



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any, Generic, cast

import nshconfig as C
import torch
from lightning.pytorch import LightningDataModule
from typing_extensions import Never, TypeVar, deprecated, override

from ..model.mixins.callback import CallbackRegistrarModuleMixin
from ..model.mixins.debug import DebugModuleMixin

THparams = TypeVar("THparams", bound=C.Config, infer_variance=True)


class LightningDataModuleBase(
    DebugModuleMixin,
    CallbackRegistrarModuleMixin,
    LightningDataModule,
    ABC,
    Generic[THparams],
):
    @property
    @override
    def hparams(self) -> THparams:  # pyright: ignore[reportIncompatibleMethodOverride]
        return cast(THparams, super().hparams)

    @property
    @override
    def hparams_initial(self) -> THparams:  # pyright: ignore[reportIncompatibleMethodOverride]
        hparams = cast(THparams, super().hparams_initial)
        return hparams

    @property
    @deprecated("Use `hparams` instead")
    def config(self):
        return cast(Never, self.hparams)

    @classmethod
    @abstractmethod
    def hparams_cls(cls) -> type[THparams]: ...

    @override
    def __init__(self, hparams: THparams | Mapping[str, Any]):
        super().__init__()

        # Validate and save hyperparameters
        hparams_cls = self.hparams_cls()
        if isinstance(hparams, Mapping):
            hparams = hparams_cls.model_validate(hparams)
        elif not isinstance(hparams, hparams_cls):
            raise TypeError(
                f"Expected hparams to be either a Mapping or an instance of {hparams_cls}, got {type(hparams)}"
            )
        hparams = hparams.model_deep_validate()
        self.save_hyperparameters(hparams)

    @override
    @classmethod
    def load_from_checkpoint(cls, *args, **kwargs) -> Never:
        raise ValueError("This method is not supported. Use `from_checkpoint` instead.")

    @classmethod
    def hparams_from_checkpoint(
        cls,
        ckpt_or_path: dict[str, Any] | str | Path,
        /,
        strict: bool | None = None,
        *,
        update_hparams: Callable[[THparams], THparams] | None = None,
        update_hparams_dict: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
    ):
        if isinstance(ckpt_or_path, dict):
            ckpt = ckpt_or_path
        else:
            ckpt = torch.load(ckpt_or_path, map_location="cpu")

        if (hparams := ckpt.get(cls.CHECKPOINT_HYPER_PARAMS_KEY)) is None:
            raise ValueError(
                f"The checkpoint does not contain hyperparameters. It must contain the key '{cls.CHECKPOINT_HYPER_PARAMS_KEY}'."
            )
        if update_hparams_dict is not None:
            hparams = update_hparams_dict(hparams)

        hparams = cls.hparams_cls().model_validate(hparams, strict=strict)
        if update_hparams is not None:
            hparams = update_hparams(hparams)

        return hparams

    @classmethod
    def from_checkpoint(
        cls,
        ckpt_or_path: dict[str, Any] | str | Path,
        /,
        strict: bool | None = None,
        map_location: torch.serialization.MAP_LOCATION = None,
        *,
        update_hparams: Callable[[THparams], THparams] | None = None,
        update_hparams_dict: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
    ):
        # Load checkpoint
        if isinstance(ckpt_or_path, Mapping):
            ckpt = ckpt_or_path
        else:
            ckpt = torch.load(ckpt_or_path, map_location=map_location)

        # Load hyperparameters from checkpoint
        hparams = cls.hparams_from_checkpoint(
            ckpt,
            strict=strict,
            update_hparams=update_hparams,
            update_hparams_dict=update_hparams_dict,
        )

        # Load datamodule from checkpoint
        datamodule = cls(hparams)
        if datamodule.__class__.__qualname__ in ckpt:
            datamodule.load_state_dict(ckpt[datamodule.__class__.__qualname__])

        return datamodule



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.log_epoch import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.log_epoch import (
    LogEpochCallbackConfig as LogEpochCallbackConfig,
)
from nshtrainer.callbacks.log_epoch import callback_registry as callback_registry

__all__ = [
    "CallbackConfigBase",
    "LogEpochCallbackConfig",
    "callback_registry",
]



from __future__ import annotations

import logging
from collections.abc import Iterable
from typing import Literal, Protocol, runtime_checkable

import torch.nn as nn
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from typing_extensions import TypeAliasType, final, override

from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class SharedParametersCallbackConfig(CallbackConfigBase):
    """A callback that allows scaling the gradients of shared parameters that
    are registered in the ``self.shared_parameters`` list of the root module.

    This is useful for models that share parameters across multiple modules and
    want to downscale the gradients of these parameters to avoid overfitting.
    """

    name: Literal["shared_parameters"] = "shared_parameters"

    @override
    def create_callbacks(self, trainer_config):
        yield SharedParametersCallback(self)


def _parameters_to_names(parameters: Iterable[nn.Parameter], model: nn.Module):
    mapping = {id(p): n for n, p in model.named_parameters()}
    return [mapping[id(p)] for p in parameters]


SharedParametersList = TypeAliasType(
    "SharedParametersList", list[tuple[nn.Parameter, int | float]]
)


@runtime_checkable
class ModuleWithSharedParameters(Protocol):
    @property
    def shared_parameters(self) -> SharedParametersList: ...


class SharedParametersCallback(Callback):
    @override
    def __init__(self, config: SharedParametersCallbackConfig):
        super().__init__()

        self.config = config
        del config

        self._warned_shared_parameters = False

    def _shared_parameters(self, pl_module: LightningModule) -> SharedParametersList:
        if not isinstance(pl_module, ModuleWithSharedParameters):
            return []

        return pl_module.shared_parameters

    @override
    def on_after_backward(self, trainer: Trainer, pl_module: LightningModule):
        if not (shared_parameters := self._shared_parameters(pl_module)):
            log.debug(
                "No shared parameters to scale, skipping SharedParametersCallback"
            )
            return

        log.debug(f"Scaling {len(shared_parameters)} shared parameters...")
        no_grad_parameters: list[nn.Parameter] = []
        for p, factor in shared_parameters:
            if not hasattr(p, "grad") or p.grad is None:
                no_grad_parameters.append(p)
                continue

            _ = p.grad.data.div_(factor)

        if no_grad_parameters and not self._warned_shared_parameters:
            no_grad_parameters_str = ", ".join(
                _parameters_to_names(no_grad_parameters, pl_module)
            )
            log.warning(
                "The following parameters were marked as shared, but had no gradients: "
                f"{no_grad_parameters_str}"
            )
            self._warned_shared_parameters = True

        log.debug(f"Done scaling shared parameters. (len={len(shared_parameters)})")



from __future__ import annotations

__codegen__ = True

from nshtrainer.lr_scheduler import (
    LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig,
)
from nshtrainer.lr_scheduler import LRSchedulerConfig as LRSchedulerConfig
from nshtrainer.lr_scheduler import LRSchedulerConfigBase as LRSchedulerConfigBase
from nshtrainer.lr_scheduler import ReduceLROnPlateauConfig as ReduceLROnPlateauConfig
from nshtrainer.lr_scheduler.base import lr_scheduler_registry as lr_scheduler_registry
from nshtrainer.lr_scheduler.linear_warmup_cosine import (
    DurationConfig as DurationConfig,
)
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import EpochsConfig as EpochsConfig
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import MetricConfig as MetricConfig

from . import base as base
from . import linear_warmup_cosine as linear_warmup_cosine
from . import reduce_lr_on_plateau as reduce_lr_on_plateau

__all__ = [
    "DurationConfig",
    "EpochsConfig",
    "LRSchedulerConfig",
    "LRSchedulerConfigBase",
    "LinearWarmupCosineDecayLRSchedulerConfig",
    "MetricConfig",
    "ReduceLROnPlateauConfig",
    "base",
    "linear_warmup_cosine",
    "lr_scheduler_registry",
    "reduce_lr_on_plateau",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.util.config.duration import DurationConfig as DurationConfig
from nshtrainer.util.config.duration import EpochsConfig as EpochsConfig
from nshtrainer.util.config.duration import StepsConfig as StepsConfig

__all__ = [
    "DurationConfig",
    "EpochsConfig",
    "StepsConfig",
]



from __future__ import annotations

from typing import Literal

from typing_extensions import override

from .base import PluginConfigBase, plugin_registry


@plugin_registry.register
class TorchSyncBatchNormPlugin(PluginConfigBase):
    """A plugin that wraps all batch normalization layers of a model with synchronization
    logic for multiprocessing.

    This plugin has no effect in single-device operation.
    """

    name: Literal["torch_sync_batchnorm"] = "torch_sync_batchnorm"

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.layer_sync import TorchSyncBatchNorm

        return TorchSyncBatchNorm()



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer.strategy import StrategyConfig as StrategyConfig
from nshtrainer.trainer.strategy import StrategyConfigBase as StrategyConfigBase

__all__ = [
    "StrategyConfig",
    "StrategyConfigBase",
]



from __future__ import annotations

from typing import Literal

from lightning.pytorch.utilities.types import LRSchedulerConfigType
from torch.optim.lr_scheduler import ReduceLROnPlateau
from typing_extensions import final, override

from ..metrics._config import MetricConfig
from ..util.config import EpochsConfig
from .base import LRSchedulerConfigBase, LRSchedulerMetadata, lr_scheduler_registry


@final
@lr_scheduler_registry.register
class ReduceLROnPlateauConfig(LRSchedulerConfigBase):
    """Reduce learning rate when a metric has stopped improving."""

    name: Literal["reduce_lr_on_plateau"] = "reduce_lr_on_plateau"

    metric: MetricConfig | None = None
    """Metric to monitor.
    If not provided, the primary metric of the runner will be used."""

    patience: int | EpochsConfig
    r"""Number of epochs with no improvement after which learning rate will be reduced."""

    factor: float
    r"""Factor by which the learning rate will be reduced. new_lr = lr * factor."""

    cooldown: int | EpochsConfig = 0
    r"""Number of epochs to wait before resuming normal operation after lr has been reduced."""

    min_lr: float | list[float] = 0.0
    r"""A scalar or a list of scalars. A lower bound on the learning rate of all param groups or each group respectively."""

    eps: float = 1.0e-8
    r"""Minimal decay applied to lr. If the difference between new and old lr is smaller than eps, the update is ignored."""

    threshold: float = 1.0e-4
    r"""Threshold for measuring the new optimum, to only focus on significant changes."""

    threshold_mode: Literal["rel", "abs"] = "rel"
    r"""One of `rel`, `abs`. In `rel` mode, dynamic_threshold = best * (1 + threshold) in 'max' mode or best * (1 - threshold) in `min` mode. In `abs` mode, dynamic_threshold = best + threshold in `max` mode or best - threshold in `min` mode. Default: 'rel'."""

    @override
    def create_scheduler_impl(
        self, optimizer, lightning_module
    ) -> LRSchedulerConfigType:
        if (metric := self.metric) is None:
            from ..trainer import Trainer

            assert isinstance(trainer := lightning_module.trainer, Trainer), (
                "The trainer must be a `nshtrainer.Trainer` instance."
            )

            assert (metric := trainer.hparams.primary_metric) is not None, (
                "Primary metric must be provided if metric is not specified."
            )

        if isinstance(patience := self.patience, EpochsConfig):
            patience = int(patience.value)

        if isinstance(cooldown := self.cooldown, EpochsConfig):
            cooldown = int(cooldown.value)

        lr_scheduler = ReduceLROnPlateau(
            optimizer,
            mode=metric.mode,
            factor=self.factor,
            patience=patience,
            threshold=self.threshold,
            threshold_mode=self.threshold_mode,
            cooldown=cooldown,
            min_lr=self.min_lr,
            eps=self.eps,
        )
        return {
            "scheduler": lr_scheduler,
            "monitor": metric.monitor,
        }

    @override
    def metadata(self) -> LRSchedulerMetadata:
        return {
            "interval": "epoch",
        }



from __future__ import annotations

import logging
from pathlib import Path
from typing import Literal

from lightning.pytorch import LightningModule, Trainer
from typing_extensions import final, override

from ..._checkpoint.metadata import CheckpointMetadata
from ...metrics._config import MetricConfig
from ..base import callback_registry
from ._base import BaseCheckpointCallbackConfig, CheckpointBase

log = logging.getLogger(__name__)


@final
@callback_registry.register
class BestCheckpointCallbackConfig(BaseCheckpointCallbackConfig):
    name: Literal["best_checkpoint"] = "best_checkpoint"

    metric: MetricConfig | None = None
    """Metric to monitor, or `None` to use the default metric."""

    throw_on_no_metric: bool = True
    """
    Whether to throw an error if no metric is provided and no primary metric is found in the root config.
    """

    @override
    def create_checkpoint(self, trainer_config, dirpath):
        # Resolve metric
        if (metric := self.metric) is None and (
            metric := trainer_config.primary_metric
        ) is None:
            error_msg = (
                "No metric provided and no primary metric found in the root config. "
                "Cannot create BestCheckpointCallback."
            )
            if self.throw_on_no_metric:
                raise ValueError(error_msg)
            else:
                log.warning(error_msg)
                return None

        return BestCheckpointCallback(self, dirpath, metric)


@final
class BestCheckpointCallback(CheckpointBase[BestCheckpointCallbackConfig]):
    @property
    def _metric_name_normalized(self):
        return self.metric.monitor.replace("/", "_").replace(" ", "_").replace(".", "_")

    @override
    def __init__(
        self,
        config: BestCheckpointCallbackConfig,
        dirpath: Path,
        metric: MetricConfig,
    ):
        self.metric = metric
        super().__init__(config, dirpath)

    @override
    def name(self):
        return f"best_{self._metric_name_normalized}"

    @override
    def default_filename(self):
        return f"epoch{{epoch}}-step{{step}}-{self._metric_name_normalized}{{{self.metric.monitor}}}"

    @override
    def topk_sort_key(self, metadata: CheckpointMetadata):
        return metadata.metrics.get(
            self.metric.monitor,
            float("-inf" if self.metric.mode == "max" else "inf"),
        )

    @override
    def topk_sort_reverse(self):
        return self.metric.mode == "max"

    # Events
    @override
    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        self.save_checkpoints(trainer)



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.gradient_skipping import (
    CallbackConfigBase as CallbackConfigBase,
)
from nshtrainer.callbacks.gradient_skipping import (
    GradientSkippingCallbackConfig as GradientSkippingCallbackConfig,
)
from nshtrainer.callbacks.gradient_skipping import (
    callback_registry as callback_registry,
)

__all__ = [
    "CallbackConfigBase",
    "GradientSkippingCallbackConfig",
    "callback_registry",
]



from __future__ import annotations

from typing import Annotated

import nshconfig as C
from typing_extensions import TypeAliasType

from .base import LRSchedulerConfigBase as LRSchedulerConfigBase
from .base import LRSchedulerMetadata as LRSchedulerMetadata
from .linear_warmup_cosine import (
    LinearWarmupCosineAnnealingLR as LinearWarmupCosineAnnealingLR,
)
from .linear_warmup_cosine import (
    LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig,
)
from .reduce_lr_on_plateau import ReduceLROnPlateau as ReduceLROnPlateau
from .reduce_lr_on_plateau import ReduceLROnPlateauConfig as ReduceLROnPlateauConfig

LRSchedulerConfig = TypeAliasType(
    "LRSchedulerConfig",
    Annotated[
        LinearWarmupCosineDecayLRSchedulerConfig | ReduceLROnPlateauConfig,
        C.Field(discriminator="name"),
    ],
)



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.debug_flag import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.debug_flag import (
    DebugFlagCallbackConfig as DebugFlagCallbackConfig,
)
from nshtrainer.callbacks.debug_flag import callback_registry as callback_registry

__all__ = [
    "CallbackConfigBase",
    "DebugFlagCallbackConfig",
    "callback_registry",
]



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import Generic, cast

import torch.nn as nn
from typing_extensions import TypeVar

TModule = TypeVar("TModule", bound=nn.Module, infer_variance=True)


class TypedModuleDict(nn.Module, Generic[TModule]):
    def __init__(
        self,
        modules: Mapping[str, TModule],
        key_prefix: str = "_typed_moduledict_",
        # we use a key prefix to avoid attribute name collisions
        # (which is a common issue in nn.ModuleDict as it uses `__setattr__` to set the modules)
    ):
        super().__init__()

        self.key_prefix = key_prefix
        self._module_dict = nn.ModuleDict(
            {self._with_prefix(k): v for k, v in modules.items()}
        )

    def _with_prefix(self, key: str) -> str:
        return f"{self.key_prefix}{key}"

    def _remove_prefix(self, key: str) -> str:
        assert key.startswith(
            self.key_prefix
        ), f"{key} does not start with {self.key_prefix}"
        return key[len(self.key_prefix) :]

    def __setitem__(self, key: str, module: TModule) -> None:
        key = self._with_prefix(key)
        return self._module_dict.__setitem__(key, module)

    def __getitem__(self, key: str) -> TModule:
        key = self._with_prefix(key)
        return self._module_dict.__getitem__(key)  # type: ignore

    def update(self, modules: Mapping[str, TModule]) -> None:
        return self._module_dict.update(
            {self._with_prefix(k): v for k, v in modules.items()}
        )

    def get(self, key: str) -> TModule | None:
        key = self._with_prefix(key)
        if (value := self._module_dict._modules.get(key)) is None:
            return None
        return cast(TModule, value)

    def keys(self) -> Iterable[str]:
        r"""Return an iterable of the ModuleDict keys."""
        return [self._remove_prefix(k) for k in self._module_dict.keys()]

    def items(self) -> Iterable[tuple[str, TModule]]:
        r"""Return an iterable of the ModuleDict key/value pairs."""
        return [
            (self._remove_prefix(k), cast(TModule, v))
            for k, v in self._module_dict.items()
        ]

    def values(self) -> Iterable[TModule]:
        r"""Return an iterable of the ModuleDict values."""
        return cast(Iterable[TModule], self._module_dict.values())



from __future__ import annotations

__codegen__ = True

from nshtrainer.profiler.advanced import (
    AdvancedProfilerConfig as AdvancedProfilerConfig,
)
from nshtrainer.profiler.advanced import BaseProfilerConfig as BaseProfilerConfig

__all__ = [
    "AdvancedProfilerConfig",
    "BaseProfilerConfig",
]



from __future__ import annotations

import logging
import math
from typing import Literal

from lightning.fabric.utilities.rank_zero import _get_rank
from lightning.pytorch import Trainer
from lightning.pytorch.callbacks import EarlyStopping as _EarlyStopping
from lightning.pytorch.utilities.rank_zero import rank_prefixed_message
from typing_extensions import final, override

from ..metrics._config import MetricConfig
from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class EarlyStoppingCallbackConfig(CallbackConfigBase):
    name: Literal["early_stopping"] = "early_stopping"

    metric: MetricConfig | None = None
    """
    The metric to monitor for early stopping.
    If None, the primary metric will be used.
    """

    patience: int
    """
    Number of epochs with no improvement after which training will be stopped.
    """

    min_delta: float = 1.0e-8
    """
    Minimum change in the monitored quantity to qualify as an improvement.
    """

    min_lr: float | None = None
    """
    Minimum learning rate. If the learning rate of the model is less than this value,
    the training will be stopped.
    """

    skip_first_n_epochs: int = 0
    """
    Number of initial epochs to skip before starting to monitor for early stopping.
    This helps avoid false early stopping when the model might temporarily perform worse
    during early training phases.
    """

    strict: bool = True
    """
    Whether to enforce that the monitored quantity must improve by at least `min_delta`
    to qualify as an improvement.
    """

    @override
    def create_callbacks(self, trainer_config):
        if (metric := self.metric) is None and (
            metric := trainer_config.primary_metric
        ) is None:
            raise ValueError(
                "Either `metric` or `trainer_config.primary_metric` must be set to use EarlyStopping."
            )

        yield EarlyStoppingCallback(self, metric)


class EarlyStoppingCallback(_EarlyStopping):
    def __init__(self, config: EarlyStoppingCallbackConfig, metric: MetricConfig):
        self.config = config
        self.metric = metric
        del config, metric

        super().__init__(
            monitor=self.metric.monitor,
            mode=self.metric.mode,
            patience=self.config.patience,
            min_delta=self.config.min_delta,
            strict=self.config.strict,
        )

    @override
    @staticmethod
    def _log_info(trainer: Trainer | None, message: str, log_rank_zero_only: bool):
        rank = _get_rank()
        if trainer is not None and trainer.world_size <= 1:
            rank = None
        message = rank_prefixed_message(message, rank)
        if rank is None or not log_rank_zero_only or rank == 0:
            log.critical(message)

    @override
    def _run_early_stopping_check(self, trainer: Trainer):
        """Checks whether the early stopping condition is met and if so tells the trainer to stop the training."""
        logs = trainer.callback_metrics

        # Disable early_stopping with fast_dev_run
        if getattr(trainer, "fast_dev_run", False):
            return

        # Skip early stopping check for the first n epochs
        if trainer.current_epoch < self.config.skip_first_n_epochs:
            if self.verbose and trainer.current_epoch == 0:
                self._log_info(
                    trainer,
                    f"Early stopping checks are disabled for the first {self.config.skip_first_n_epochs} epochs",
                    self.log_rank_zero_only,
                )
            return

        should_stop, reason = False, None

        if not should_stop:
            should_stop, reason = self._evaluate_stopping_criteria_min_lr(trainer)

        # If metric present
        if not should_stop and self._validate_condition_metric(logs):
            current = logs[self.monitor].squeeze()
            should_stop, reason = self._evaluate_stopping_criteria(current)

        # stop every ddp process if any world process decides to stop
        should_stop = trainer.strategy.reduce_boolean_decision(should_stop, all=False)
        trainer.should_stop = trainer.should_stop or should_stop
        if should_stop:
            self.stopped_epoch = trainer.current_epoch
        if reason and self.verbose:
            self._log_info(trainer, reason, self.log_rank_zero_only)

    def _evaluate_stopping_criteria_min_lr(
        self, trainer: Trainer
    ) -> tuple[bool, str | None]:
        if self.config.min_lr is None:
            return False, None

        # Get the maximum LR across all param groups in all optimizers
        model_max_lr = max(
            [
                param_group["lr"]
                for optimizer in trainer.optimizers
                for param_group in optimizer.param_groups
            ]
        )
        if not isinstance(model_max_lr, float) or not math.isfinite(model_max_lr):
            return False, None

        # If the maximum LR is less than the minimum LR, stop training
        if model_max_lr >= self.config.min_lr:
            return False, None

        return True, (
            "Stopping threshold reached: "
            f"The maximum LR of the model across all param groups is {model_max_lr:.2e} "
            f"which is less than the minimum LR {self.config.min_lr:.2e}"
        )

    def on_early_stopping(self, trainer: Trainer):
        pass



from __future__ import annotations

__codegen__ = True

from nshtrainer import MetricConfig as MetricConfig
from nshtrainer import TrainerConfig as TrainerConfig
from nshtrainer._checkpoint.metadata import CheckpointMetadata as CheckpointMetadata
from nshtrainer._directory import DirectoryConfig as DirectoryConfig
from nshtrainer._hf_hub import CallbackConfigBase as CallbackConfigBase
from nshtrainer._hf_hub import (
    HuggingFaceHubAutoCreateConfig as HuggingFaceHubAutoCreateConfig,
)
from nshtrainer._hf_hub import HuggingFaceHubConfig as HuggingFaceHubConfig
from nshtrainer._hf_hub import callback_registry as callback_registry
from nshtrainer.callbacks import ActSaveConfig as ActSaveConfig
from nshtrainer.callbacks import (
    BestCheckpointCallbackConfig as BestCheckpointCallbackConfig,
)
from nshtrainer.callbacks import CallbackConfig as CallbackConfig
from nshtrainer.callbacks import DebugFlagCallbackConfig as DebugFlagCallbackConfig
from nshtrainer.callbacks import (
    DirectorySetupCallbackConfig as DirectorySetupCallbackConfig,
)
from nshtrainer.callbacks import (
    DistributedPredictionWriterConfig as DistributedPredictionWriterConfig,
)
from nshtrainer.callbacks import (
    EarlyStoppingCallbackConfig as EarlyStoppingCallbackConfig,
)
from nshtrainer.callbacks import EMACallbackConfig as EMACallbackConfig
from nshtrainer.callbacks import EpochTimerCallbackConfig as EpochTimerCallbackConfig
from nshtrainer.callbacks import (
    FiniteChecksCallbackConfig as FiniteChecksCallbackConfig,
)
from nshtrainer.callbacks import (
    GradientSkippingCallbackConfig as GradientSkippingCallbackConfig,
)
from nshtrainer.callbacks import (
    LastCheckpointCallbackConfig as LastCheckpointCallbackConfig,
)
from nshtrainer.callbacks import LearningRateMonitorConfig as LearningRateMonitorConfig
from nshtrainer.callbacks import LogEpochCallbackConfig as LogEpochCallbackConfig
from nshtrainer.callbacks import (
    MetricValidationCallbackConfig as MetricValidationCallbackConfig,
)
from nshtrainer.callbacks import NormLoggingCallbackConfig as NormLoggingCallbackConfig
from nshtrainer.callbacks import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from nshtrainer.callbacks import (
    PrintTableMetricsCallbackConfig as PrintTableMetricsCallbackConfig,
)
from nshtrainer.callbacks import (
    RLPSanityChecksCallbackConfig as RLPSanityChecksCallbackConfig,
)
from nshtrainer.callbacks import (
    SharedParametersCallbackConfig as SharedParametersCallbackConfig,
)
from nshtrainer.callbacks import (
    WandbUploadCodeCallbackConfig as WandbUploadCodeCallbackConfig,
)
from nshtrainer.callbacks import WandbWatchCallbackConfig as WandbWatchCallbackConfig
from nshtrainer.callbacks.checkpoint._base import (
    BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig,
)
from nshtrainer.loggers import ActSaveLoggerConfig as ActSaveLoggerConfig
from nshtrainer.loggers import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.loggers import LoggerConfig as LoggerConfig
from nshtrainer.loggers import LoggerConfigBase as LoggerConfigBase
from nshtrainer.loggers import TensorboardLoggerConfig as TensorboardLoggerConfig
from nshtrainer.loggers import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.loggers import logger_registry as logger_registry
from nshtrainer.lr_scheduler import (
    LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig,
)
from nshtrainer.lr_scheduler import LRSchedulerConfig as LRSchedulerConfig
from nshtrainer.lr_scheduler import LRSchedulerConfigBase as LRSchedulerConfigBase
from nshtrainer.lr_scheduler import ReduceLROnPlateauConfig as ReduceLROnPlateauConfig
from nshtrainer.lr_scheduler.base import lr_scheduler_registry as lr_scheduler_registry
from nshtrainer.nn import ELUNonlinearityConfig as ELUNonlinearityConfig
from nshtrainer.nn import GELUNonlinearityConfig as GELUNonlinearityConfig
from nshtrainer.nn import LeakyReLUNonlinearityConfig as LeakyReLUNonlinearityConfig
from nshtrainer.nn import MishNonlinearityConfig as MishNonlinearityConfig
from nshtrainer.nn import MLPConfig as MLPConfig
from nshtrainer.nn import NonlinearityConfig as NonlinearityConfig
from nshtrainer.nn import NonlinearityConfigBase as NonlinearityConfigBase
from nshtrainer.nn import PReLUConfig as PReLUConfig
from nshtrainer.nn import ReLUNonlinearityConfig as ReLUNonlinearityConfig
from nshtrainer.nn import RNGConfig as RNGConfig
from nshtrainer.nn import SigmoidNonlinearityConfig as SigmoidNonlinearityConfig
from nshtrainer.nn import SiLUNonlinearityConfig as SiLUNonlinearityConfig
from nshtrainer.nn import SoftmaxNonlinearityConfig as SoftmaxNonlinearityConfig
from nshtrainer.nn import SoftplusNonlinearityConfig as SoftplusNonlinearityConfig
from nshtrainer.nn import SoftsignNonlinearityConfig as SoftsignNonlinearityConfig
from nshtrainer.nn import SwishNonlinearityConfig as SwishNonlinearityConfig
from nshtrainer.nn import TanhNonlinearityConfig as TanhNonlinearityConfig
from nshtrainer.nn.nonlinearity import (
    SwiGLUNonlinearityConfig as SwiGLUNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import nonlinearity_registry as nonlinearity_registry
from nshtrainer.optimizer import AdadeltaConfig as AdadeltaConfig
from nshtrainer.optimizer import AdafactorConfig as AdafactorConfig
from nshtrainer.optimizer import AdagradConfig as AdagradConfig
from nshtrainer.optimizer import AdamaxConfig as AdamaxConfig
from nshtrainer.optimizer import AdamConfig as AdamConfig
from nshtrainer.optimizer import AdamWConfig as AdamWConfig
from nshtrainer.optimizer import ASGDConfig as ASGDConfig
from nshtrainer.optimizer import NAdamConfig as NAdamConfig
from nshtrainer.optimizer import OptimizerConfig as OptimizerConfig
from nshtrainer.optimizer import OptimizerConfigBase as OptimizerConfigBase
from nshtrainer.optimizer import RAdamConfig as RAdamConfig
from nshtrainer.optimizer import RMSpropConfig as RMSpropConfig
from nshtrainer.optimizer import RpropConfig as RpropConfig
from nshtrainer.optimizer import SGDConfig as SGDConfig
from nshtrainer.optimizer import Union as Union
from nshtrainer.optimizer import optimizer_registry as optimizer_registry
from nshtrainer.profiler import AdvancedProfilerConfig as AdvancedProfilerConfig
from nshtrainer.profiler import BaseProfilerConfig as BaseProfilerConfig
from nshtrainer.profiler import ProfilerConfig as ProfilerConfig
from nshtrainer.profiler import PyTorchProfilerConfig as PyTorchProfilerConfig
from nshtrainer.profiler import SimpleProfilerConfig as SimpleProfilerConfig
from nshtrainer.trainer import accelerator_registry as accelerator_registry
from nshtrainer.trainer import plugin_registry as plugin_registry
from nshtrainer.trainer._config import AcceleratorConfig as AcceleratorConfig
from nshtrainer.trainer._config import (
    CheckpointCallbackConfig as CheckpointCallbackConfig,
)
from nshtrainer.trainer._config import CheckpointSavingConfig as CheckpointSavingConfig
from nshtrainer.trainer._config import EnvironmentConfig as EnvironmentConfig
from nshtrainer.trainer._config import GradientClippingConfig as GradientClippingConfig
from nshtrainer.trainer._config import SanityCheckingConfig as SanityCheckingConfig
from nshtrainer.trainer._config import StrategyConfig as StrategyConfig
from nshtrainer.trainer.accelerator import CPUAcceleratorConfig as CPUAcceleratorConfig
from nshtrainer.trainer.accelerator import (
    CUDAAcceleratorConfig as CUDAAcceleratorConfig,
)
from nshtrainer.trainer.accelerator import MPSAcceleratorConfig as MPSAcceleratorConfig
from nshtrainer.trainer.accelerator import XLAAcceleratorConfig as XLAAcceleratorConfig
from nshtrainer.trainer.plugin import PluginConfig as PluginConfig
from nshtrainer.trainer.plugin import PluginConfigBase as PluginConfigBase
from nshtrainer.trainer.plugin.environment import (
    KubeflowEnvironmentPlugin as KubeflowEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    LightningEnvironmentPlugin as LightningEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    LSFEnvironmentPlugin as LSFEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    MPIEnvironmentPlugin as MPIEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    SLURMEnvironmentPlugin as SLURMEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    TorchElasticEnvironmentPlugin as TorchElasticEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    XLAEnvironmentPlugin as XLAEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.io import (
    AsyncCheckpointIOPlugin as AsyncCheckpointIOPlugin,
)
from nshtrainer.trainer.plugin.io import (
    TorchCheckpointIOPlugin as TorchCheckpointIOPlugin,
)
from nshtrainer.trainer.plugin.io import XLACheckpointIOPlugin as XLACheckpointIOPlugin
from nshtrainer.trainer.plugin.layer_sync import (
    TorchSyncBatchNormPlugin as TorchSyncBatchNormPlugin,
)
from nshtrainer.trainer.plugin.precision import (
    BitsandbytesPluginConfig as BitsandbytesPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    DeepSpeedPluginConfig as DeepSpeedPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    DoublePrecisionPluginConfig as DoublePrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    FSDPPrecisionPluginConfig as FSDPPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    HalfPrecisionPluginConfig as HalfPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    MixedPrecisionPluginConfig as MixedPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    TransformerEnginePluginConfig as TransformerEnginePluginConfig,
)
from nshtrainer.trainer.plugin.precision import XLAPluginConfig as XLAPluginConfig
from nshtrainer.trainer.trainer import AcceleratorConfigBase as AcceleratorConfigBase
from nshtrainer.trainer.trainer import StrategyConfigBase as StrategyConfigBase
from nshtrainer.util._environment_info import (
    EnvironmentClassInformationConfig as EnvironmentClassInformationConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentCUDAConfig as EnvironmentCUDAConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentGPUConfig as EnvironmentGPUConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentHardwareConfig as EnvironmentHardwareConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentLinuxEnvironmentConfig as EnvironmentLinuxEnvironmentConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentLSFInformationConfig as EnvironmentLSFInformationConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentPackageConfig as EnvironmentPackageConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentSLURMInformationConfig as EnvironmentSLURMInformationConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentSnapshotConfig as EnvironmentSnapshotConfig,
)
from nshtrainer.util._environment_info import GitRepositoryConfig as GitRepositoryConfig
from nshtrainer.util.config import DTypeConfig as DTypeConfig
from nshtrainer.util.config import DurationConfig as DurationConfig
from nshtrainer.util.config import EpochsConfig as EpochsConfig
from nshtrainer.util.config import StepsConfig as StepsConfig

from . import _checkpoint as _checkpoint
from . import _directory as _directory
from . import _hf_hub as _hf_hub
from . import callbacks as callbacks
from . import loggers as loggers
from . import lr_scheduler as lr_scheduler
from . import metrics as metrics
from . import nn as nn
from . import optimizer as optimizer
from . import profiler as profiler
from . import trainer as trainer
from . import util as util

__all__ = [
    "ASGDConfig",
    "AcceleratorConfig",
    "AcceleratorConfigBase",
    "ActSaveConfig",
    "ActSaveLoggerConfig",
    "AdadeltaConfig",
    "AdafactorConfig",
    "AdagradConfig",
    "AdamConfig",
    "AdamWConfig",
    "AdamaxConfig",
    "AdvancedProfilerConfig",
    "AsyncCheckpointIOPlugin",
    "BaseCheckpointCallbackConfig",
    "BaseProfilerConfig",
    "BestCheckpointCallbackConfig",
    "BitsandbytesPluginConfig",
    "CPUAcceleratorConfig",
    "CSVLoggerConfig",
    "CUDAAcceleratorConfig",
    "CallbackConfig",
    "CallbackConfigBase",
    "CheckpointCallbackConfig",
    "CheckpointMetadata",
    "CheckpointSavingConfig",
    "DTypeConfig",
    "DebugFlagCallbackConfig",
    "DeepSpeedPluginConfig",
    "DirectoryConfig",
    "DirectorySetupCallbackConfig",
    "DistributedPredictionWriterConfig",
    "DoublePrecisionPluginConfig",
    "DurationConfig",
    "ELUNonlinearityConfig",
    "EMACallbackConfig",
    "EarlyStoppingCallbackConfig",
    "EnvironmentCUDAConfig",
    "EnvironmentClassInformationConfig",
    "EnvironmentConfig",
    "EnvironmentGPUConfig",
    "EnvironmentHardwareConfig",
    "EnvironmentLSFInformationConfig",
    "EnvironmentLinuxEnvironmentConfig",
    "EnvironmentPackageConfig",
    "EnvironmentSLURMInformationConfig",
    "EnvironmentSnapshotConfig",
    "EpochTimerCallbackConfig",
    "EpochsConfig",
    "FSDPPrecisionPluginConfig",
    "FiniteChecksCallbackConfig",
    "GELUNonlinearityConfig",
    "GitRepositoryConfig",
    "GradientClippingConfig",
    "GradientSkippingCallbackConfig",
    "HalfPrecisionPluginConfig",
    "HuggingFaceHubAutoCreateConfig",
    "HuggingFaceHubConfig",
    "KubeflowEnvironmentPlugin",
    "LRSchedulerConfig",
    "LRSchedulerConfigBase",
    "LSFEnvironmentPlugin",
    "LastCheckpointCallbackConfig",
    "LeakyReLUNonlinearityConfig",
    "LearningRateMonitorConfig",
    "LightningEnvironmentPlugin",
    "LinearWarmupCosineDecayLRSchedulerConfig",
    "LogEpochCallbackConfig",
    "LoggerConfig",
    "LoggerConfigBase",
    "MLPConfig",
    "MPIEnvironmentPlugin",
    "MPSAcceleratorConfig",
    "MetricConfig",
    "MetricValidationCallbackConfig",
    "MishNonlinearityConfig",
    "MixedPrecisionPluginConfig",
    "NAdamConfig",
    "NonlinearityConfig",
    "NonlinearityConfigBase",
    "NormLoggingCallbackConfig",
    "OnExceptionCheckpointCallbackConfig",
    "OptimizerConfig",
    "OptimizerConfigBase",
    "PReLUConfig",
    "PluginConfig",
    "PluginConfigBase",
    "PrintTableMetricsCallbackConfig",
    "ProfilerConfig",
    "PyTorchProfilerConfig",
    "RAdamConfig",
    "RLPSanityChecksCallbackConfig",
    "RMSpropConfig",
    "RNGConfig",
    "ReLUNonlinearityConfig",
    "ReduceLROnPlateauConfig",
    "RpropConfig",
    "SGDConfig",
    "SLURMEnvironmentPlugin",
    "SanityCheckingConfig",
    "SharedParametersCallbackConfig",
    "SiLUNonlinearityConfig",
    "SigmoidNonlinearityConfig",
    "SimpleProfilerConfig",
    "SoftmaxNonlinearityConfig",
    "SoftplusNonlinearityConfig",
    "SoftsignNonlinearityConfig",
    "StepsConfig",
    "StrategyConfig",
    "StrategyConfigBase",
    "SwiGLUNonlinearityConfig",
    "SwishNonlinearityConfig",
    "TanhNonlinearityConfig",
    "TensorboardLoggerConfig",
    "TorchCheckpointIOPlugin",
    "TorchElasticEnvironmentPlugin",
    "TorchSyncBatchNormPlugin",
    "TrainerConfig",
    "TransformerEnginePluginConfig",
    "Union",
    "WandbLoggerConfig",
    "WandbUploadCodeCallbackConfig",
    "WandbWatchCallbackConfig",
    "XLAAcceleratorConfig",
    "XLACheckpointIOPlugin",
    "XLAEnvironmentPlugin",
    "XLAPluginConfig",
    "_checkpoint",
    "_directory",
    "_hf_hub",
    "accelerator_registry",
    "callback_registry",
    "callbacks",
    "logger_registry",
    "loggers",
    "lr_scheduler",
    "lr_scheduler_registry",
    "metrics",
    "nn",
    "nonlinearity_registry",
    "optimizer",
    "optimizer_registry",
    "plugin_registry",
    "profiler",
    "trainer",
    "util",
]



from __future__ import annotations

from typing import Annotated

from typing_extensions import TypeAliasType

from . import environment as environment
from . import io as io
from . import layer_sync as layer_sync
from . import precision as precision
from .base import Plugin as Plugin
from .base import PluginConfigBase as PluginConfigBase
from .base import plugin_registry as plugin_registry

PluginConfig = TypeAliasType(
    "PluginConfig",
    Annotated[PluginConfigBase, plugin_registry.DynamicResolution()],
)



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.checkpoint.last_checkpoint import (
    BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint.last_checkpoint import (
    CheckpointMetadata as CheckpointMetadata,
)
from nshtrainer.callbacks.checkpoint.last_checkpoint import (
    LastCheckpointCallbackConfig as LastCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint.last_checkpoint import (
    callback_registry as callback_registry,
)

__all__ = [
    "BaseCheckpointCallbackConfig",
    "CheckpointMetadata",
    "LastCheckpointCallbackConfig",
    "callback_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer._checkpoint.metadata import CheckpointMetadata as CheckpointMetadata
from nshtrainer._checkpoint.metadata import EnvironmentConfig as EnvironmentConfig

from . import metadata as metadata

__all__ = [
    "CheckpointMetadata",
    "EnvironmentConfig",
    "metadata",
]



from __future__ import annotations

import contextlib
import datetime
import logging
import os
from pathlib import Path
from typing import Any, Literal

from lightning.pytorch import Trainer as LightningTrainer
from lightning.pytorch.callbacks import OnExceptionCheckpoint as _OnExceptionCheckpoint
from typing_extensions import final, override

from ..base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@contextlib.contextmanager
def _monkey_patch_disable_barrier(trainer: LightningTrainer):
    """
    Monkey-patch the strategy instance to make the barrier operation a no-op.

    We do this because `save_checkpoint` calls `barrier`. This is okay in most
    cases, but when we want to save a checkpoint in the case of an exception,
    `barrier` causes a deadlock. So we monkey-patch the strategy instance to
    make the barrier operation a no-op.
    """

    # We monkey-patch the barrier method to do nothing.
    original_barrier = trainer.strategy.barrier

    def new_barrier(*args, **kwargs):
        log.warning("Monkey-patched no-op barrier.")
        pass

    trainer.strategy.barrier = new_barrier
    log.warning("Monkey-patched barrier to no-op.")

    try:
        yield
    finally:
        trainer.strategy.barrier = original_barrier
        log.warning("Reverted monkey-patched barrier.")


@final
@callback_registry.register
class OnExceptionCheckpointCallbackConfig(CallbackConfigBase):
    name: Literal["on_exception_checkpoint"] = "on_exception_checkpoint"

    dirpath: str | Path | None = None
    """Directory path to save the checkpoint file."""

    filename: str | None = None
    """Checkpoint filename. This must not include the extension. If `None`, `on_exception_{id}_{timestamp}` is used."""

    @override
    def create_callbacks(self, trainer_config):
        dirpath = self.dirpath or trainer_config.directory.resolve_subdirectory(
            trainer_config.id, "checkpoint"
        )

        if not (filename := self.filename):
            filename = f"on_exception_{trainer_config.id}"
        yield OnExceptionCheckpointCallback(
            self, dirpath=Path(dirpath), filename=filename
        )


class OnExceptionCheckpointCallback(_OnExceptionCheckpoint):
    @override
    def __init__(
        self,
        config: OnExceptionCheckpointCallbackConfig,
        dirpath: Path,
        filename: str,
    ):
        self.config = config
        del config

        dirpath = dirpath / "on_exception"
        dirpath.mkdir(parents=True, exist_ok=True)

        super().__init__(dirpath, filename)

    @property
    @override
    def ckpt_path(self) -> str:
        ckpt_path = super().ckpt_path

        # Remve the extension and add the current timestamp
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        ckpt_path, ext = os.path.splitext(ckpt_path)
        return f"{ckpt_path}_{timestamp}{ext}"

    @override
    def on_exception(self, trainer: LightningTrainer, *args: Any, **kwargs: Any):
        # Monkey-patch the strategy instance to make the barrier operation a no-op.
        # We do this because `save_checkpoint` calls `barrier`. This is okay in most
        #   cases, but when we want to save a checkpoint in the case of an exception,
        #   `barrier` causes a deadlock. So we monkey-patch the strategy instance to
        #   make the barrier operation a no-op.
        with _monkey_patch_disable_barrier(trainer):
            return super().on_exception(trainer, *args, **kwargs)



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer.plugin.io import (
    AsyncCheckpointIOPlugin as AsyncCheckpointIOPlugin,
)
from nshtrainer.trainer.plugin.io import PluginConfigBase as PluginConfigBase
from nshtrainer.trainer.plugin.io import (
    TorchCheckpointIOPlugin as TorchCheckpointIOPlugin,
)
from nshtrainer.trainer.plugin.io import XLACheckpointIOPlugin as XLACheckpointIOPlugin
from nshtrainer.trainer.plugin.io import plugin_registry as plugin_registry

__all__ = [
    "AsyncCheckpointIOPlugin",
    "PluginConfigBase",
    "TorchCheckpointIOPlugin",
    "XLACheckpointIOPlugin",
    "plugin_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.loggers.actsave import ActSaveLoggerConfig as ActSaveLoggerConfig
from nshtrainer.loggers.actsave import LoggerConfigBase as LoggerConfigBase
from nshtrainer.loggers.actsave import logger_registry as logger_registry

__all__ = [
    "ActSaveLoggerConfig",
    "LoggerConfigBase",
    "logger_registry",
]



from __future__ import annotations

import builtins
from typing import Any, Literal

import nshconfig as C


class MetricConfig(C.Config):
    monitor: str
    """The name of the metric to monitor."""

    mode: Literal["min", "max"]
    """
    The mode of the primary metric:
    - "min" for metrics that should be minimized (e.g., loss)
    - "max" for metrics that should be maximized (e.g., accuracy)
    """

    @property
    def best(self):
        return builtins.min if self.mode == "min" else builtins.max

    def is_better(self, a: Any, b: Any):
        return self.best(a, b) == a



from __future__ import annotations

__codegen__ = True

from nshtrainer._directory import DirectoryConfig as DirectoryConfig
from nshtrainer._directory import (
    DirectorySetupCallbackConfig as DirectorySetupCallbackConfig,
)
from nshtrainer._directory import LoggerConfig as LoggerConfig

__all__ = [
    "DirectoryConfig",
    "DirectorySetupCallbackConfig",
    "LoggerConfig",
]



from __future__ import annotations

import logging
from pathlib import Path
from typing import Literal

from typing_extensions import final, override

from .._callback import NTCallbackBase
from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class DirectorySetupCallbackConfig(CallbackConfigBase):
    name: Literal["directory_setup"] = "directory_setup"

    enabled: bool = True
    """Whether to enable the directory setup callback."""

    create_symlink_to_nshrunner_root: bool = True
    """Should we create a symlink to the root folder for the Runner (if we're in one)?"""

    def __bool__(self):
        return self.enabled

    @override
    def create_callbacks(self, trainer_config):
        if not self:
            return

        yield DirectorySetupCallback(self)


def _create_symlink_to_nshrunner(base_dir: Path):
    try:
        import nshrunner as nr
    except ImportError:
        log.info("nshrunner is not installed. Skipping symlink creation to nshrunner.")
        return

    # Check if we are in a nshrunner session
    if (session := nr.Session.from_current_session()) is None:
        log.info("No current nshrunner session found. Skipping symlink creation.")
        return

    session_dir = session.session_dir
    if not session_dir.exists() or not session_dir.is_dir():
        log.warning(
            f"nshrunner's session_dir is not a valid directory: {session_dir}. "
            "Skipping symlink creation."
        )
        return

    # Create the symlink
    symlink_path = base_dir / "nshrunner"
    if symlink_path.exists(follow_symlinks=False):
        # If it already points to the correct directory, we're done
        if symlink_path.resolve() == session_dir.resolve():
            return

        # Otherwise, we should log a warning and remove the existing symlink
        log.warning(
            f"A symlink pointing to {symlink_path.resolve()} already exists at {symlink_path}. "
            "Removing the existing symlink."
        )
        symlink_path.unlink()

    symlink_path.symlink_to(session_dir, target_is_directory=True)


class DirectorySetupCallback(NTCallbackBase):
    @override
    def __init__(self, config: DirectorySetupCallbackConfig):
        super().__init__()

        self.config = config
        del config

    @override
    def setup(self, trainer, pl_module, stage):
        super().setup(trainer, pl_module, stage)

        # Create a symlink to the root folder for the Runner
        if self.config.create_symlink_to_nshrunner_root:
            base_dir = trainer.hparams.directory.resolve_run_root_directory(
                trainer.hparams.id
            )
            _create_symlink_to_nshrunner(base_dir)



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.directory_setup import (
    CallbackConfigBase as CallbackConfigBase,
)
from nshtrainer.callbacks.directory_setup import (
    DirectorySetupCallbackConfig as DirectorySetupCallbackConfig,
)
from nshtrainer.callbacks.directory_setup import callback_registry as callback_registry

__all__ = [
    "CallbackConfigBase",
    "DirectorySetupCallbackConfig",
    "callback_registry",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Annotated, Literal

import nshconfig as C
from lightning.pytorch.accelerators import Accelerator
from typing_extensions import TypeAliasType, override

if TYPE_CHECKING:
    from ._config import TrainerConfig

AcceleratorLiteral = TypeAliasType(
    "AcceleratorLiteral", Literal["cpu", "gpu", "tpu", "ipu", "hpu", "mps", "auto"]
)


class AcceleratorConfigBase(C.Config, ABC):
    @abstractmethod
    def create_accelerator(self, trainer_config: "TrainerConfig") -> Accelerator: ...


accelerator_registry = C.Registry(AcceleratorConfigBase, discriminator="name")

AcceleratorConfig = TypeAliasType(
    "AcceleratorConfig",
    Annotated[AcceleratorConfigBase, accelerator_registry.DynamicResolution()],
)


@accelerator_registry.register
class CPUAcceleratorConfig(AcceleratorConfigBase):
    name: Literal["cpu"] = "cpu"

    """Accelerator for CPU devices."""

    @override
    def create_accelerator(self, trainer_config) -> Accelerator:
        from lightning.pytorch.accelerators.cpu import CPUAccelerator

        return CPUAccelerator()


@accelerator_registry.register
class CUDAAcceleratorConfig(AcceleratorConfigBase):
    name: Literal["gpu"] = "gpu"

    """Accelerator for NVIDIA CUDA devices."""

    @override
    def create_accelerator(self, trainer_config) -> Accelerator:
        from lightning.pytorch.accelerators.cuda import CUDAAccelerator

        return CUDAAccelerator()


@accelerator_registry.register
class MPSAcceleratorConfig(AcceleratorConfigBase):
    name: Literal["mps"] = "mps"

    """Accelerator for Metal Apple Silicon GPU devices.

    .. warning::  Use of this accelerator beyond import and instantiation is experimental.
    """

    @override
    def create_accelerator(self, trainer_config) -> Accelerator:
        from lightning.pytorch.accelerators.mps import MPSAccelerator

        return MPSAccelerator()


@accelerator_registry.register
class XLAAcceleratorConfig(AcceleratorConfigBase):
    name: Literal["tpu"] = "tpu"

    """Accelerator for XLA devices, normally TPUs.

    .. warning::  Use of this accelerator beyond import and instantiation is experimental.
    """

    @override
    def create_accelerator(self, trainer_config) -> Accelerator:
        from lightning.pytorch.accelerators.xla import XLAAccelerator

        return XLAAccelerator()



from __future__ import annotations

import logging
import os
from collections.abc import Callable, Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast, overload

import torch
from lightning.fabric.plugins.environments.lsf import LSFEnvironment
from lightning.fabric.plugins.environments.slurm import SLURMEnvironment
from lightning.fabric.plugins.precision.precision import _PRECISION_INPUT
from lightning.pytorch import LightningDataModule, LightningModule
from lightning.pytorch import Trainer as LightningTrainer
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.profilers import Profiler
from lightning.pytorch.trainer.states import TrainerFn
from lightning.pytorch.utilities.types import (
    _EVALUATE_OUTPUT,
    _PREDICT_OUTPUT,
    EVAL_DATALOADERS,
)
from typing_extensions import Never, Unpack, assert_never, deprecated, override

from .._checkpoint.metadata import write_checkpoint_metadata
from ..callbacks.base import resolve_all_callbacks
from ..callbacks.distributed_prediction_writer import (
    DistributedPredictionWriter,
    DistributedPredictionWriterConfig,
)
from ..util._environment_info import EnvironmentConfig
from ..util.bf16 import is_bf16_supported_no_emulation
from ._config import LightningTrainerKwargs, TrainerConfig
from ._distributed_prediction_result import DistributedPredictionResult
from ._log_hparams import patch_log_hparams_function
from ._runtime_callback import RuntimeTrackerCallback, Stage
from .accelerator import AcceleratorConfigBase
from .signal_connector import _SignalConnector
from .strategy import StrategyConfigBase

log = logging.getLogger(__name__)


patch_log_hparams_function()


class Trainer(LightningTrainer):
    CHECKPOINT_HYPER_PARAMS_KEY = "trainer_hyper_parameters"

    @property
    def hparams(self) -> TrainerConfig:
        """The collection of hyperparameters saved with :meth:`save_hyperparameters`. It is mutable by the user. For
        the frozen set of initial hyperparameters, use :attr:`hparams_initial`.

        Returns:
            Mutable hyperparameters dictionary

        """
        return self._hparams

    @property
    @deprecated("Use `hparams` instead")
    def config(self):
        return cast(Never, self.hparams)

    @classmethod
    def hparams_cls(cls):
        return TrainerConfig

    @classmethod
    def _pre_init(cls, hparams: TrainerConfig):
        if (precision := hparams.set_float32_matmul_precision) is not None:
            torch.set_float32_matmul_precision(precision)

    @classmethod
    def _update_kwargs(
        cls,
        hparams: TrainerConfig,
        kwargs_ctor: LightningTrainerKwargs,
    ):
        kwargs: LightningTrainerKwargs = {
            "deterministic": hparams.deterministic,
            "fast_dev_run": hparams.fast_dev_run,
            "max_epochs": hparams.max_epochs,
            "min_epochs": hparams.min_epochs,
            "max_steps": hparams.max_steps,
            "min_steps": hparams.min_steps,
            "max_time": hparams.max_time,
            "limit_train_batches": hparams.limit_train_batches,
            "limit_val_batches": hparams.limit_val_batches,
            "limit_test_batches": hparams.limit_test_batches,
            "limit_predict_batches": hparams.limit_predict_batches,
            "overfit_batches": hparams.overfit_batches,
            "val_check_interval": hparams.val_check_interval,
            "num_sanity_val_steps": hparams.num_sanity_val_steps,
            "log_every_n_steps": hparams.log_every_n_steps,
            "inference_mode": hparams.inference_mode,
            "callbacks": [],
            "plugins": [],
            "logger": [],
            # Moved to `lightning_kwargs`:
            # "enable_checkpointing": hparams.enable_checkpointing,
            # "accelerator": hparams.accelerator,
            # "strategy": hparams.strategy,
            # "num_nodes": hparams.num_nodes,
            # "precision": hparams.precision,
            # "logger": hparams.logging.enabled,
            # "log_every_n_steps": hparams.log_every_n_steps,
            # "enable_progress_bar": hparams.enable_progress_bar,
            # "enable_model_summary": hparams.enable_model_summary,
            # "accumulate_grad_batches": hparams.accumulate_grad_batches,
            # "benchmark": hparams.benchmark,
            # "use_distributed_sampler": hparams.use_distributed_sampler,
            # "detect_anomaly": hparams.detect_anomaly,
            # "barebones": hparams.barebones,
            # "plugins": hparams.plugins,
            # "sync_batchnorm": hparams.sync_batchnorm,
            # "reload_dataloaders_every_n_epochs": hparams.reload_dataloaders_every_n_epochs,
        }

        def _update_key(key: str, new_value: Any):
            # First, check to see if the key is already in the kwargs.
            if key not in kwargs:
                kwargs[key] = new_value
                return

            # If the key is already in the kwargs, then we check the type:
            # - If the type is a sequence, then we extend the sequence.
            # - Otherwise, we just update the value but warn the user.

            match existing_value := kwargs[key]:
                case Sequence() as existing_value:
                    # Make sure value is a sequence too
                    if not isinstance(new_value, Sequence):
                        new_value = [new_value]
                    kwargs[key] = [*existing_value, *new_value]
                case _:
                    log.warning(
                        f"Trainer.__init__: Overwriting existing value {existing_value=} with {new_value=} for key {key=}."
                    )
                    kwargs[key] = new_value

        def _update_kwargs(**update: Unpack[LightningTrainerKwargs]):
            for key, value in update.items():
                _update_key(key, value)

        # Set `barebones`
        if hparams.barebones:
            _update_kwargs(barebones=True)

        # Set `default_root_dir` if `auto_set_default_root_dir` is enabled.
        if hparams.auto_set_default_root_dir:
            if kwargs.get("default_root_dir"):
                raise ValueError(
                    "You have set `hparams.default_root_dir`. "
                    "But we are trying to set it automatically. "
                    "Please use `hparams.directory.base` rather than `hparams.default_root_dir`. "
                    "If you want to set it manually, please set `hparams.auto_set_default_root_dir=False`."
                )

            _update_kwargs(
                default_root_dir=hparams.directory.resolve_run_root_directory(
                    hparams.id
                )
            )

        if (devices_input := hparams.devices) is not None:
            match devices_input:
                case "all":
                    devices = -1
                case "auto":
                    devices = "auto"
                case Sequence():
                    devices = list(devices_input)
                case _:
                    raise ValueError(f"Invalid value for devices={devices_input}.")

            _update_kwargs(devices=devices)

        if (use_distributed_sampler := hparams.use_distributed_sampler) is not None:
            _update_kwargs(use_distributed_sampler=use_distributed_sampler)

        if (accelerator := hparams.accelerator) is not None:
            if isinstance(accelerator, AcceleratorConfigBase):
                accelerator = accelerator.create_accelerator(hparams)
            _update_kwargs(accelerator=accelerator)

        if (strategy := hparams.strategy) is not None:
            if isinstance(strategy, StrategyConfigBase):
                strategy = strategy.create_strategy(hparams)
            _update_kwargs(strategy=strategy)

        if (precision := hparams.precision) is not None:
            resolved_precision: _PRECISION_INPUT
            match precision:
                case "64-true" | "32-true" | "bf16-mixed":
                    resolved_precision = precision
                case "fp16-mixed":
                    resolved_precision = "16-mixed"
                case "16-mixed-auto":
                    try:
                        resolved_precision = (
                            "bf16-mixed"
                            if is_bf16_supported_no_emulation()
                            else "16-mixed"
                        )
                    except BaseException:
                        resolved_precision = "16-mixed"
                        log.warning(
                            "Failed to detect bfloat16 support. Falling back to 16-mixed."
                        )

                    log.critical(
                        f"Auto-resolving {precision=} to {resolved_precision=}."
                    )
                case _:
                    assert_never(precision)

            _update_kwargs(precision=resolved_precision)

        if (detect_anomaly := hparams.detect_anomaly) is not None:
            _update_kwargs(detect_anomaly=detect_anomaly)

        if (
            grad_clip_config := hparams.gradient_clipping
        ) is not None and grad_clip_config.enabled:
            # kwargs["gradient_clip_algorithm"] = grad_clip_config.algorithm
            # kwargs["gradient_clip_val"] = grad_clip_config.value
            _update_kwargs(
                gradient_clip_algorithm=grad_clip_config.algorithm,
                gradient_clip_val=grad_clip_config.value,
            )

        if profiler_config := hparams.profiler:
            if (profiler := profiler_config.create_profiler(hparams)) is None:
                log.warning(f"Profiler hparams {profiler_config=} returned None.")
            # Make sure that the profiler is an instance of `Profiler`.
            elif not isinstance(profiler, Profiler):
                raise ValueError(f"{profiler=} is not an instance of `{Profiler}`.")
            # Otherwise, if the profiler is a string (e.g., "simpe", "advanced", "pytorch"),
            #   then we just pass it through.
            else:
                _update_kwargs(profiler=profiler)

        if callbacks := resolve_all_callbacks(hparams):
            _update_kwargs(callbacks=callbacks)

        if plugin_configs := hparams.plugins:
            _update_kwargs(
                plugins=[
                    plugin_config.create_plugin(hparams)
                    for plugin_config in plugin_configs
                ]
            )

        _update_kwargs(
            logger=[
                logger
                for logger_config in hparams._nshtrainer_all_logger_configs()
                if logger_config is not None
                and (logger := logger_config.create_logger(hparams)) is not None
            ]
        )

        if hparams.auto_determine_num_nodes:
            # When num_nodes is auto, we need to detect the number of nodes.
            if SLURMEnvironment.detect():
                if (num_nodes := os.environ.get("SLURM_NNODES")) is not None:
                    num_nodes = int(num_nodes)
                    log.critical(f"SLURM detected with {num_nodes=}.")
                    _update_kwargs(num_nodes=num_nodes)
                else:
                    log.critical(
                        "SLURM detected, but SLURM_NNODES not found. "
                        "We'll continue without setting num_nodes, but this may cause issues."
                    )

            elif LSFEnvironment.detect():
                num_nodes = LSFEnvironment().world_size()
                log.critical(f"LSF detected with {num_nodes=}.")
                _update_kwargs(num_nodes=num_nodes)
            else:
                log.info(
                    "hparams.auto_determine_num_nodes ignored because no SLURM or LSF detected."
                )

        # Update the kwargs with the additional trainer kwargs
        _update_kwargs(**cast(Any, hparams.additional_lightning_kwargs))
        _update_kwargs(**hparams.lightning_kwargs)
        _update_kwargs(**kwargs_ctor)

        # Handle barebones mode
        if kwargs.get("barebones"):
            # Remove the logger if it's an empty list
            if (logger := kwargs.get("logger")) is not None and not logger:
                kwargs["logger"] = None

        return kwargs

    if TYPE_CHECKING:
        callbacks: list[Callback]

    @override
    def __init__(
        self,
        hparams: TrainerConfig | Mapping[str, Any],
        /,
        **kwargs: Unpack[LightningTrainerKwargs],
    ):
        # Validate the hparams.
        hparams_cls = Trainer.hparams_cls()
        if isinstance(hparams, Mapping):
            hparams = hparams_cls.model_validate(hparams)
        elif not isinstance(hparams, hparams_cls):
            raise ValueError(
                f"Trainer hparams must either be an instance of {hparams_cls} or a mapping. "
                f"Got {type(hparams)=} instead."
            )
        hparams = hparams.model_deep_validate()
        hparams._nshtrainer_validate_before_run()

        self._pre_init(hparams)

        kwargs = self._update_kwargs(hparams, kwargs)
        log.critical(f"LightningTrainer.__init__ with {kwargs=}.")

        self._hparams = hparams
        self.debug = self.hparams.debug

        experimental_profiler = None
        if hparams.experimental_barebones_profiler_enabled:
            experimental_profiler = kwargs.pop("profiler", None)
            log.warning(
                "Barebones profiler is enabled. This is an experimental feature and may not work as expected."
            )

        experimental_barebones_progress_bar = None
        if hparams.experimental_barebones_progress_bar_enabled:
            experimental_barebones_progress_bar = kwargs.pop(
                "enable_progress_bar", True
            )

        super().__init__(**kwargs)

        # Set up the profiler again
        if experimental_profiler is not None:
            from lightning.pytorch.trainer import setup

            setup._init_profiler(self, experimental_profiler)

        # Set up the progress bar again
        if experimental_barebones_progress_bar is not None:
            self._callback_connector._configure_progress_bar(
                experimental_barebones_progress_bar
            )

        # Add our own start time callback to measure the start time.
        self.callbacks.append(RuntimeTrackerCallback())

        # Replace the signal connector with our own.
        self._signal_connector = _SignalConnector(self)

        # Print out the log dir, so that we can easily find it in the logs.
        if log_dir := self.log_dir:
            log_dir = str(Path(log_dir).resolve())
        log.critical(f"LightningTrainer log directory: {self.log_dir}.")

        # Set the checkpoint
        if (ckpt_path := hparams.ckpt_path) is not None:
            self.ckpt_path = str(Path(ckpt_path).resolve().absolute())

    def __runtime_tracker(self):
        return next(
            (
                callback
                for callback in self.callbacks
                if isinstance(callback, RuntimeTrackerCallback)
            ),
            None,
        )

    def __current_stage(self) -> Stage:
        match self.state.fn:
            case None:
                raise ValueError(
                    "Trainer state function is not set. "
                    "You must call `fit`, `validate`, `test`, or `predict`, "
                    "or explicitly provide a stage."
                )
            case TrainerFn.FITTING:
                return "train"
            case TrainerFn.VALIDATING:
                return "validate"
            case TrainerFn.TESTING:
                return "test"
            case TrainerFn.PREDICTING:
                return "predict"
            case _:
                assert_never(self.state.fn)

    def start_time(self, stage: Stage | None = None):
        """Return the start time of the run"""
        if (tracker := self.__runtime_tracker()) is None:
            raise ValueError(
                "RuntimeTrackerCallback is not set. Cannot get start time."
            )
        if stage is None:
            stage = self.__current_stage()

        return tracker.start_time(stage)

    def end_time(self, stage: Stage | None = None):
        """Return the end time of the run"""
        if (tracker := self.__runtime_tracker()) is None:
            raise ValueError(
                "RuntimeTrackerCallback is not set. Cannot get start time."
            )
        if stage is None:
            stage = self.__current_stage()

        return tracker.end_time(stage)

    def time_elapsed(self, stage: Stage | None = None):
        """Return the time elapsed for the run"""
        if (tracker := self.__runtime_tracker()) is None:
            raise ValueError(
                "RuntimeTrackerCallback is not set. Cannot get start time."
            )
        if stage is None:
            stage = self.__current_stage()

        return tracker.time_elapsed(stage)

    @override
    def _run(
        self, model: LightningModule, ckpt_path: str | Path | None = None
    ) -> _EVALUATE_OUTPUT | _PREDICT_OUTPUT | None:
        """Lightning doesn't support gradient clipping with manual optimization.
        We patch the `Trainer._run` method to throw if gradient clipping is enabled
        and `model.automatic_optimization` is False.
        """
        # Save the current environment information
        datamodule = getattr(self, "datamodule", None)
        self.hparams.environment = EnvironmentConfig.from_current_environment(
            self.hparams, model, datamodule
        )

        # If gradient clipping is enabled, then we need to make sure that
        # `model.automatic_optimization` is enabled. Otherwise, gradient clipping
        # is not actually going to do anything, as we expect the user to manually
        # call `optimizer.step()` and `optimizer.zero_grad()`.
        if not model.automatic_optimization and (
            self.gradient_clip_val is not None
            or self.gradient_clip_algorithm is not None
        ):
            raise ValueError(
                "Automatic gradient clipping is not supported with manual optimization. "
                f"Please set {model.__class__.__name__}.automatic_optimization to True "
                "or disable automatic gradient clipping. "
            )

        return super()._run(model, ckpt_path)

    @override
    def save_checkpoint(
        self,
        filepath: str | Path,
        weights_only: bool = False,
        storage_options: Any | None = None,
    ):
        filepath = Path(filepath)

        if self.model is None:
            raise AttributeError(
                "Saving a checkpoint is only possible if a model is attached to the Trainer. Did you call"
                " `Trainer.save_checkpoint()` before calling `Trainer.{fit,validate,test,predict}`?"
            )
        with self.profiler.profile("save_checkpoint"):  # type: ignore
            checkpoint = self._checkpoint_connector.dump_checkpoint(weights_only)
            # Update the checkpoint for the trainer hyperparameters
            checkpoint[self.CHECKPOINT_HYPER_PARAMS_KEY] = self.hparams.model_dump(
                mode="json"
            )
            self.strategy.save_checkpoint(
                checkpoint, filepath, storage_options=storage_options
            )
            self.strategy.barrier("Trainer.save_checkpoint")

        # Save the checkpoint metadata
        metadata_path = None
        if self.hparams.save_checkpoint_metadata and self.is_global_zero:
            # Generate the metadata and write to disk
            metadata_path = write_checkpoint_metadata(self, filepath)

        # Call the `on_checkpoint_saved` method on all callbacks
        from .. import _callback

        _callback._call_on_checkpoint_saved(self, filepath, metadata_path)

    @classmethod
    def hparams_from_checkpoint(
        cls,
        ckpt_or_path: dict[str, Any] | str | Path,
        /,
        strict: bool | None = None,
        *,
        update_hparams: Callable[[TrainerConfig], TrainerConfig] | None = None,
        update_hparams_dict: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
    ):
        if isinstance(ckpt_or_path, dict):
            ckpt = ckpt_or_path
        else:
            ckpt = torch.load(ckpt_or_path, map_location="cpu")

        if (hparams := ckpt.get(cls.CHECKPOINT_HYPER_PARAMS_KEY)) is None:
            raise ValueError(
                f"The checkpoint does not contain hyperparameters. It must contain the key '{cls.CHECKPOINT_HYPER_PARAMS_KEY}'."
            )
        if update_hparams_dict is not None:
            hparams = update_hparams_dict(hparams)

        hparams = cls.hparams_cls().model_validate(hparams, strict=strict)
        if update_hparams is not None:
            hparams = update_hparams(hparams)

        return hparams

    @classmethod
    def from_checkpoint(
        cls,
        path: str | Path,
        strict: bool | None = None,
        *,
        update_hparams: Callable[[TrainerConfig], TrainerConfig] | None = None,
        update_hparams_dict: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
    ):
        hparams = cls.hparams_from_checkpoint(
            path,
            strict=strict,
            update_hparams=update_hparams,
            update_hparams_dict=update_hparams_dict,
        )
        return cls(hparams)

    @overload
    def distributed_predict(
        self,
        model: LightningModule | None = None,
        dataloaders: EVAL_DATALOADERS | LightningDataModule | None = None,
        datamodule: LightningDataModule | None = None,
        ckpt_path: str | Path | None = None,
        *,
        config: DistributedPredictionWriterConfig,
    ) -> DistributedPredictionResult: ...

    @overload
    def distributed_predict(
        self,
        model: LightningModule | None = None,
        dataloaders: EVAL_DATALOADERS | LightningDataModule | None = None,
        datamodule: LightningDataModule | None = None,
        ckpt_path: str | Path | None = None,
        *,
        dirpath: Path | None = None,
        move_to_cpu_on_save: bool = True,
        save_raw: bool = True,
        save_processed: bool = True,
    ) -> DistributedPredictionResult: ...

    def distributed_predict(
        self,
        model: LightningModule | None = None,
        dataloaders: EVAL_DATALOADERS | LightningDataModule | None = None,
        datamodule: LightningDataModule | None = None,
        ckpt_path: str | Path | None = None,
        *,
        config: DistributedPredictionWriterConfig | None = None,
        dirpath: Path | None = None,
        move_to_cpu_on_save: bool = True,
        save_raw: bool = True,
        save_processed: bool = True,
    ) -> DistributedPredictionResult:
        if config is None:
            config = DistributedPredictionWriterConfig(
                dirpath=dirpath,
                move_to_cpu_on_save=move_to_cpu_on_save,
                save_raw=save_raw,
                save_processed=save_processed,
            )

        # Remove any DistributedPredictionWriter callbacks that are already set
        # and add the new one.
        callbacks = self.callbacks.copy()
        callbacks = [
            callback
            for callback in callbacks
            if not isinstance(callback, DistributedPredictionWriter)
        ]
        writer_callbacks = list(config.create_callbacks(self.hparams))
        assert len(writer_callbacks) == 1
        callback = writer_callbacks[0]
        callbacks.append(callback)
        self.callbacks = self._callback_connector._reorder_callbacks(callbacks)

        self.predict(
            model,
            dataloaders,
            datamodule,
            return_predictions=False,
            ckpt_path=ckpt_path,
        )

        # Wait for all processes to finish
        self.strategy.barrier("Trainer.distributed_predict")

        # Return an object that contains information about the predictions
        return DistributedPredictionResult(root_dir=callback.output_dir)



from __future__ import annotations

__codegen__ = True

from nshtrainer.nn import ELUNonlinearityConfig as ELUNonlinearityConfig
from nshtrainer.nn import GELUNonlinearityConfig as GELUNonlinearityConfig
from nshtrainer.nn import LeakyReLUNonlinearityConfig as LeakyReLUNonlinearityConfig
from nshtrainer.nn import MishNonlinearityConfig as MishNonlinearityConfig
from nshtrainer.nn import MLPConfig as MLPConfig
from nshtrainer.nn import NonlinearityConfig as NonlinearityConfig
from nshtrainer.nn import NonlinearityConfigBase as NonlinearityConfigBase
from nshtrainer.nn import PReLUConfig as PReLUConfig
from nshtrainer.nn import ReLUNonlinearityConfig as ReLUNonlinearityConfig
from nshtrainer.nn import RNGConfig as RNGConfig
from nshtrainer.nn import SigmoidNonlinearityConfig as SigmoidNonlinearityConfig
from nshtrainer.nn import SiLUNonlinearityConfig as SiLUNonlinearityConfig
from nshtrainer.nn import SoftmaxNonlinearityConfig as SoftmaxNonlinearityConfig
from nshtrainer.nn import SoftplusNonlinearityConfig as SoftplusNonlinearityConfig
from nshtrainer.nn import SoftsignNonlinearityConfig as SoftsignNonlinearityConfig
from nshtrainer.nn import SwishNonlinearityConfig as SwishNonlinearityConfig
from nshtrainer.nn import TanhNonlinearityConfig as TanhNonlinearityConfig
from nshtrainer.nn.nonlinearity import (
    SwiGLUNonlinearityConfig as SwiGLUNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import nonlinearity_registry as nonlinearity_registry

from . import mlp as mlp
from . import nonlinearity as nonlinearity
from . import rng as rng

__all__ = [
    "ELUNonlinearityConfig",
    "GELUNonlinearityConfig",
    "LeakyReLUNonlinearityConfig",
    "MLPConfig",
    "MishNonlinearityConfig",
    "NonlinearityConfig",
    "NonlinearityConfigBase",
    "PReLUConfig",
    "RNGConfig",
    "ReLUNonlinearityConfig",
    "SiLUNonlinearityConfig",
    "SigmoidNonlinearityConfig",
    "SoftmaxNonlinearityConfig",
    "SoftplusNonlinearityConfig",
    "SoftsignNonlinearityConfig",
    "SwiGLUNonlinearityConfig",
    "SwishNonlinearityConfig",
    "TanhNonlinearityConfig",
    "mlp",
    "nonlinearity",
    "nonlinearity_registry",
    "rng",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer import TrainerConfig as TrainerConfig
from nshtrainer.trainer import accelerator_registry as accelerator_registry
from nshtrainer.trainer import callback_registry as callback_registry
from nshtrainer.trainer import plugin_registry as plugin_registry
from nshtrainer.trainer._config import AcceleratorConfig as AcceleratorConfig
from nshtrainer.trainer._config import ActSaveLoggerConfig as ActSaveLoggerConfig
from nshtrainer.trainer._config import (
    BestCheckpointCallbackConfig as BestCheckpointCallbackConfig,
)
from nshtrainer.trainer._config import CallbackConfig as CallbackConfig
from nshtrainer.trainer._config import CallbackConfigBase as CallbackConfigBase
from nshtrainer.trainer._config import (
    CheckpointCallbackConfig as CheckpointCallbackConfig,
)
from nshtrainer.trainer._config import CheckpointSavingConfig as CheckpointSavingConfig
from nshtrainer.trainer._config import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.trainer._config import (
    DebugFlagCallbackConfig as DebugFlagCallbackConfig,
)
from nshtrainer.trainer._config import DirectoryConfig as DirectoryConfig
from nshtrainer.trainer._config import (
    EarlyStoppingCallbackConfig as EarlyStoppingCallbackConfig,
)
from nshtrainer.trainer._config import EnvironmentConfig as EnvironmentConfig
from nshtrainer.trainer._config import GradientClippingConfig as GradientClippingConfig
from nshtrainer.trainer._config import HuggingFaceHubConfig as HuggingFaceHubConfig
from nshtrainer.trainer._config import (
    LastCheckpointCallbackConfig as LastCheckpointCallbackConfig,
)
from nshtrainer.trainer._config import (
    LearningRateMonitorConfig as LearningRateMonitorConfig,
)
from nshtrainer.trainer._config import LogEpochCallbackConfig as LogEpochCallbackConfig
from nshtrainer.trainer._config import LoggerConfig as LoggerConfig
from nshtrainer.trainer._config import LoggerConfigBase as LoggerConfigBase
from nshtrainer.trainer._config import MetricConfig as MetricConfig
from nshtrainer.trainer._config import (
    MetricValidationCallbackConfig as MetricValidationCallbackConfig,
)
from nshtrainer.trainer._config import (
    NormLoggingCallbackConfig as NormLoggingCallbackConfig,
)
from nshtrainer.trainer._config import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from nshtrainer.trainer._config import ProfilerConfig as ProfilerConfig
from nshtrainer.trainer._config import (
    RLPSanityChecksCallbackConfig as RLPSanityChecksCallbackConfig,
)
from nshtrainer.trainer._config import SanityCheckingConfig as SanityCheckingConfig
from nshtrainer.trainer._config import (
    SharedParametersCallbackConfig as SharedParametersCallbackConfig,
)
from nshtrainer.trainer._config import StrategyConfig as StrategyConfig
from nshtrainer.trainer._config import (
    TensorboardLoggerConfig as TensorboardLoggerConfig,
)
from nshtrainer.trainer._config import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.trainer.accelerator import CPUAcceleratorConfig as CPUAcceleratorConfig
from nshtrainer.trainer.accelerator import (
    CUDAAcceleratorConfig as CUDAAcceleratorConfig,
)
from nshtrainer.trainer.accelerator import MPSAcceleratorConfig as MPSAcceleratorConfig
from nshtrainer.trainer.accelerator import XLAAcceleratorConfig as XLAAcceleratorConfig
from nshtrainer.trainer.plugin import PluginConfig as PluginConfig
from nshtrainer.trainer.plugin import PluginConfigBase as PluginConfigBase
from nshtrainer.trainer.plugin.environment import (
    KubeflowEnvironmentPlugin as KubeflowEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    LightningEnvironmentPlugin as LightningEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    LSFEnvironmentPlugin as LSFEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    MPIEnvironmentPlugin as MPIEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    SLURMEnvironmentPlugin as SLURMEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    TorchElasticEnvironmentPlugin as TorchElasticEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    XLAEnvironmentPlugin as XLAEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.io import (
    AsyncCheckpointIOPlugin as AsyncCheckpointIOPlugin,
)
from nshtrainer.trainer.plugin.io import (
    TorchCheckpointIOPlugin as TorchCheckpointIOPlugin,
)
from nshtrainer.trainer.plugin.io import XLACheckpointIOPlugin as XLACheckpointIOPlugin
from nshtrainer.trainer.plugin.layer_sync import (
    TorchSyncBatchNormPlugin as TorchSyncBatchNormPlugin,
)
from nshtrainer.trainer.plugin.precision import (
    BitsandbytesPluginConfig as BitsandbytesPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    DeepSpeedPluginConfig as DeepSpeedPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    DoublePrecisionPluginConfig as DoublePrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import DTypeConfig as DTypeConfig
from nshtrainer.trainer.plugin.precision import (
    FSDPPrecisionPluginConfig as FSDPPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    HalfPrecisionPluginConfig as HalfPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    MixedPrecisionPluginConfig as MixedPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    TransformerEnginePluginConfig as TransformerEnginePluginConfig,
)
from nshtrainer.trainer.plugin.precision import XLAPluginConfig as XLAPluginConfig
from nshtrainer.trainer.trainer import AcceleratorConfigBase as AcceleratorConfigBase
from nshtrainer.trainer.trainer import (
    DistributedPredictionWriterConfig as DistributedPredictionWriterConfig,
)
from nshtrainer.trainer.trainer import StrategyConfigBase as StrategyConfigBase

from . import _config as _config
from . import accelerator as accelerator
from . import plugin as plugin
from . import strategy as strategy
from . import trainer as trainer

__all__ = [
    "AcceleratorConfig",
    "AcceleratorConfigBase",
    "ActSaveLoggerConfig",
    "AsyncCheckpointIOPlugin",
    "BestCheckpointCallbackConfig",
    "BitsandbytesPluginConfig",
    "CPUAcceleratorConfig",
    "CSVLoggerConfig",
    "CUDAAcceleratorConfig",
    "CallbackConfig",
    "CallbackConfigBase",
    "CheckpointCallbackConfig",
    "CheckpointSavingConfig",
    "DTypeConfig",
    "DebugFlagCallbackConfig",
    "DeepSpeedPluginConfig",
    "DirectoryConfig",
    "DistributedPredictionWriterConfig",
    "DoublePrecisionPluginConfig",
    "EarlyStoppingCallbackConfig",
    "EnvironmentConfig",
    "FSDPPrecisionPluginConfig",
    "GradientClippingConfig",
    "HalfPrecisionPluginConfig",
    "HuggingFaceHubConfig",
    "KubeflowEnvironmentPlugin",
    "LSFEnvironmentPlugin",
    "LastCheckpointCallbackConfig",
    "LearningRateMonitorConfig",
    "LightningEnvironmentPlugin",
    "LogEpochCallbackConfig",
    "LoggerConfig",
    "LoggerConfigBase",
    "MPIEnvironmentPlugin",
    "MPSAcceleratorConfig",
    "MetricConfig",
    "MetricValidationCallbackConfig",
    "MixedPrecisionPluginConfig",
    "NormLoggingCallbackConfig",
    "OnExceptionCheckpointCallbackConfig",
    "PluginConfig",
    "PluginConfigBase",
    "ProfilerConfig",
    "RLPSanityChecksCallbackConfig",
    "SLURMEnvironmentPlugin",
    "SanityCheckingConfig",
    "SharedParametersCallbackConfig",
    "StrategyConfig",
    "StrategyConfigBase",
    "TensorboardLoggerConfig",
    "TorchCheckpointIOPlugin",
    "TorchElasticEnvironmentPlugin",
    "TorchSyncBatchNormPlugin",
    "TrainerConfig",
    "TransformerEnginePluginConfig",
    "WandbLoggerConfig",
    "XLAAcceleratorConfig",
    "XLACheckpointIOPlugin",
    "XLAEnvironmentPlugin",
    "XLAPluginConfig",
    "_config",
    "accelerator",
    "accelerator_registry",
    "callback_registry",
    "plugin",
    "plugin_registry",
    "strategy",
    "trainer",
]



from __future__ import annotations

import importlib.metadata
import logging
from typing import TYPE_CHECKING, Literal

import nshconfig as C
from lightning.pytorch import Callback, LightningModule, Trainer
from packaging import version
from typing_extensions import assert_never, final, override

from ..callbacks.base import CallbackConfigBase
from ..callbacks.wandb_upload_code import WandbUploadCodeCallbackConfig
from ..callbacks.wandb_watch import WandbWatchCallbackConfig
from .base import LoggerConfigBase, logger_registry

if TYPE_CHECKING:
    from ..trainer._config import TrainerConfig


log = logging.getLogger(__name__)


def _project_name(
    trainer_config: TrainerConfig,
    default_project: str = "lightning_logs",
):
    # If the config has a project name, use that.
    if project := trainer_config.project:
        return project

    # Otherwise, we should use the name of the module that the config is defined in,
    #   if we can find it.
    # If this isn't in a module, use the default project name.
    if not (module := trainer_config.__module__):
        return default_project

    # If the module is a package, use the package name.
    if not (module := module.split(".", maxsplit=1)[0].strip()):
        return default_project

    return module


def _wandb_available():
    try:
        from lightning.pytorch.loggers.wandb import _WANDB_AVAILABLE

        if not _WANDB_AVAILABLE:
            log.warning("WandB not found. Disabling WandbLogger.")
            return False
        return True
    except ImportError:
        return False


class FinishWandbOnTeardownCallback(Callback):
    @override
    def teardown(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        stage: str,
    ):
        try:
            import wandb  # type: ignore
        except ImportError:
            return

        if wandb.run is None:
            return

        wandb.finish()


@final
@logger_registry.register
class WandbLoggerConfig(CallbackConfigBase, LoggerConfigBase):
    name: Literal["wandb"] = "wandb"

    enabled: bool = C.Field(default_factory=lambda: _wandb_available())
    """Enable WandB logging."""

    priority: int = 2
    """Priority of the logger. Higher priority loggers are created first,
    and the highest priority logger is the "main" logger for PyTorch Lightning."""

    project: str | None = None
    """WandB project name to use for the logger. If None, will use the root config's project name."""

    log_model: Literal["all", "latest", "none"] | bool = False
    """
    Whether to log the model checkpoints to wandb.
    Valid values are:
    - "all": Log all checkpoints.
    - "latest" or True: Log only the latest checkpoint.
    - "none" or False: Do not log any checkpoints
    """

    log_code: WandbUploadCodeCallbackConfig | None = WandbUploadCodeCallbackConfig()
    """WandB code upload configuration. Used to upload code to WandB."""

    watch: WandbWatchCallbackConfig | None = WandbWatchCallbackConfig()
    """WandB model watch configuration. Used to log model architecture, gradients, and parameters."""

    offline: bool = False
    """Whether to run WandB in offline mode."""

    use_wandb_core: bool = True
    """Whether to use the new `wandb-core` backend for WandB.
    `wandb-core` is a new backend for WandB that is faster and more efficient than the old backend.
    """

    def offline_(self, value: bool = True):
        self.offline = value
        return self

    def core_(self, value: bool = True):
        self.use_wandb_core = value
        return self

    @property
    def _lightning_log_model(self) -> Literal["all"] | bool:
        match self.log_model:
            case "all":
                return "all"
            case "latest" | True:
                return True
            case "none" | False:
                return False
            case _:
                assert_never(self.log_model)

    @override
    def create_logger(self, trainer_config):
        if not self.enabled:
            return None

        # If `wandb-core` is enabled, we should use the new backend.
        if self.use_wandb_core:
            try:
                import wandb  # type: ignore

                # The minimum version that supports the new backend is 0.17.5
                wandb_version = version.parse(importlib.metadata.version("wandb"))
                if wandb_version < version.parse("0.17.5"):
                    log.warning(
                        "The version of WandB installed does not support the `wandb-core` backend "
                        f"(expected version >= 0.17.5, found version {wandb.__version__}). "
                        "Please either upgrade to a newer version of WandB or disable the `use_wandb_core` option."
                    )
                # W&B versions 0.18.0 use wandb-core by default
                elif wandb_version < version.parse("0.18.0"):
                    wandb.require("core")  # type: ignore
                    log.critical("Using the `wandb-core` backend for WandB.")
            except ImportError:
                pass
        else:
            # W&B versions 0.18.0 use wandb-core by default,
            #   so if `use_wandb_core` is False, we should use the old backend
            #   explicitly.
            wandb_version = version.parse(importlib.metadata.version("wandb"))
            if wandb_version >= version.parse("0.18.0"):
                log.warning(
                    "Explicitly using the old backend for WandB. "
                    "If you want to use the new `wandb-core` backend, set `use_wandb_core=True`."
                )
                try:
                    import wandb  # type: ignore

                    wandb.require("legacy-service")  # type: ignore
                except ImportError:
                    pass

        from lightning.pytorch.loggers.wandb import WandbLogger

        save_dir = trainer_config.directory._resolve_log_directory_for_logger(
            trainer_config.id,
            self,
        )
        return WandbLogger(
            save_dir=save_dir,
            project=self.project or _project_name(trainer_config),
            name=trainer_config.full_name,
            version=trainer_config.id,
            log_model=self._lightning_log_model,
            notes=(
                "\n".join(f"- {note}" for note in trainer_config.notes)
                if trainer_config.notes
                else None
            ),
            tags=trainer_config.tags,
            offline=self.offline,
        )

    @override
    def create_callbacks(self, trainer_config):
        yield FinishWandbOnTeardownCallback()

        if self.watch:
            yield from self.watch.create_callbacks(trainer_config)

        if self.log_code:
            yield from self.log_code.create_callbacks(trainer_config)



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.norm_logging import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.norm_logging import (
    NormLoggingCallbackConfig as NormLoggingCallbackConfig,
)
from nshtrainer.callbacks.norm_logging import callback_registry as callback_registry

__all__ = [
    "CallbackConfigBase",
    "NormLoggingCallbackConfig",
    "callback_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.profiler.simple import BaseProfilerConfig as BaseProfilerConfig
from nshtrainer.profiler.simple import SimpleProfilerConfig as SimpleProfilerConfig

__all__ = [
    "BaseProfilerConfig",
    "SimpleProfilerConfig",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.timer import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.timer import (
    EpochTimerCallbackConfig as EpochTimerCallbackConfig,
)
from nshtrainer.callbacks.timer import callback_registry as callback_registry

__all__ = [
    "CallbackConfigBase",
    "EpochTimerCallbackConfig",
    "callback_registry",
]



from __future__ import annotations

import math
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Literal

import nshconfig as C
from lightning.pytorch import LightningModule
from lightning.pytorch.utilities.types import (
    LRSchedulerConfigType,
    LRSchedulerTypeUnion,
)
from torch.optim import Optimizer
from typing_extensions import Never, NotRequired, TypedDict


class LRSchedulerMetadata(TypedDict):
    interval: Literal["epoch", "step"]
    """Interval to update the learning rate."""

    name: NotRequired[str | None]
    """Name of the learning rate scheduler. Default is `None`."""

    frequency: NotRequired[int]
    """Frequency to update the learning rate. Default is `1`."""

    reduce_on_plateau: NotRequired[bool]
    """Whether to reduce the learning rate on plateau. Default is `False`."""

    monitor: NotRequired[str | None]
    """Value to monitor for reducing the learning rate on plateau. Required if `reduce_on_plateau` is `True`.
    Default is `None`."""

    strict: NotRequired[bool]
    """Whether to enforce that the monitor exists for reducing the learning rate on plateau. Default is `True`."""


class LRSchedulerConfigBase(C.Config, ABC):
    @abstractmethod
    def metadata(self) -> LRSchedulerMetadata: ...

    @abstractmethod
    def create_scheduler_impl(
        self, optimizer: Optimizer, lightning_module: LightningModule
    ) -> LRSchedulerTypeUnion | LRSchedulerConfigType: ...

    def create_scheduler(
        self,
        optimizer: Optimizer,
        lightning_module: LightningModule,
        lr: Never
        | None = None,  # Backward compatibility, should be removed in the future
    ) -> LRSchedulerConfigType:
        # Create the scheduler.
        scheduler = self.create_scheduler_impl(optimizer, lightning_module)

        # If the scheduler is not a `LRSchedulerConfigType`, then make it one.
        if not isinstance(scheduler, Mapping):
            scheduler = LRSchedulerConfigType(scheduler=scheduler)

        # Update the scheduler config with the metadata (if not already present).
        metadata = self.metadata()
        # - `interval` has to be present.
        if scheduler.get("interval") is None:
            scheduler["interval"] = metadata["interval"]
        # - `name`
        if scheduler.get("name") is None and "name" in metadata:
            scheduler["name"] = metadata["name"]
        # - `frequency`
        if scheduler.get("frequency") is None and "frequency" in metadata:
            scheduler["frequency"] = metadata["frequency"]
        # - `reduce_on_plateau`
        if (
            scheduler.get("reduce_on_plateau") is None
            and "reduce_on_plateau" in metadata
        ):
            scheduler["reduce_on_plateau"] = metadata["reduce_on_plateau"]
        # - `monitor`
        if scheduler.get("monitor") is None and "monitor" in metadata:
            scheduler["monitor"] = metadata["monitor"]
        # - `strict`
        if scheduler.get("strict") is None and "strict" in metadata:
            scheduler["strict"] = metadata["strict"]  # type: ignore

        return scheduler

    def compute_num_steps_per_epoch(self, lightning_module: LightningModule) -> int:
        trainer = lightning_module.trainer
        # Use the Lightning trainer to convert the epoch-based values to step-based values
        _ = trainer.estimated_stepping_batches
        # ^ This is a hack to trigger the computation of the estimated stepping batches
        #   and make sure that the `trainer.num_training_batches` attribute is set.
        return math.ceil(trainer.num_training_batches / trainer.accumulate_grad_batches)


lr_scheduler_registry = C.Registry(LRSchedulerConfigBase, discriminator="name")



from __future__ import annotations

__codegen__ = True

from nshtrainer.callbacks.checkpoint import (
    BestCheckpointCallbackConfig as BestCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint import (
    LastCheckpointCallbackConfig as LastCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint._base import (
    BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint._base import (
    CallbackConfigBase as CallbackConfigBase,
)
from nshtrainer.callbacks.checkpoint._base import (
    CheckpointMetadata as CheckpointMetadata,
)
from nshtrainer.callbacks.checkpoint.best_checkpoint import MetricConfig as MetricConfig
from nshtrainer.callbacks.checkpoint.best_checkpoint import (
    callback_registry as callback_registry,
)

from . import _base as _base
from . import best_checkpoint as best_checkpoint
from . import last_checkpoint as last_checkpoint
from . import on_exception_checkpoint as on_exception_checkpoint

__all__ = [
    "BaseCheckpointCallbackConfig",
    "BestCheckpointCallbackConfig",
    "CallbackConfigBase",
    "CheckpointMetadata",
    "LastCheckpointCallbackConfig",
    "MetricConfig",
    "OnExceptionCheckpointCallbackConfig",
    "_base",
    "best_checkpoint",
    "callback_registry",
    "last_checkpoint",
    "on_exception_checkpoint",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.util.config import DTypeConfig as DTypeConfig
from nshtrainer.util.config import DurationConfig as DurationConfig
from nshtrainer.util.config import EpochsConfig as EpochsConfig
from nshtrainer.util.config import StepsConfig as StepsConfig

from . import dtype as dtype
from . import duration as duration

__all__ = [
    "DTypeConfig",
    "DurationConfig",
    "EpochsConfig",
    "StepsConfig",
    "dtype",
    "duration",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.optimizer import AdadeltaConfig as AdadeltaConfig
from nshtrainer.optimizer import AdafactorConfig as AdafactorConfig
from nshtrainer.optimizer import AdagradConfig as AdagradConfig
from nshtrainer.optimizer import AdamaxConfig as AdamaxConfig
from nshtrainer.optimizer import AdamConfig as AdamConfig
from nshtrainer.optimizer import AdamWConfig as AdamWConfig
from nshtrainer.optimizer import ASGDConfig as ASGDConfig
from nshtrainer.optimizer import NAdamConfig as NAdamConfig
from nshtrainer.optimizer import OptimizerConfig as OptimizerConfig
from nshtrainer.optimizer import OptimizerConfigBase as OptimizerConfigBase
from nshtrainer.optimizer import RAdamConfig as RAdamConfig
from nshtrainer.optimizer import RMSpropConfig as RMSpropConfig
from nshtrainer.optimizer import RpropConfig as RpropConfig
from nshtrainer.optimizer import SGDConfig as SGDConfig
from nshtrainer.optimizer import Union as Union
from nshtrainer.optimizer import optimizer_registry as optimizer_registry

__all__ = [
    "ASGDConfig",
    "AdadeltaConfig",
    "AdafactorConfig",
    "AdagradConfig",
    "AdamConfig",
    "AdamWConfig",
    "AdamaxConfig",
    "NAdamConfig",
    "OptimizerConfig",
    "OptimizerConfigBase",
    "RAdamConfig",
    "RMSpropConfig",
    "RpropConfig",
    "SGDConfig",
    "Union",
    "optimizer_registry",
]



from __future__ import annotations

import logging
from pathlib import Path

import nshconfig as C

from .callbacks.directory_setup import DirectorySetupCallbackConfig
from .loggers import LoggerConfig

log = logging.getLogger(__name__)


class DirectoryConfig(C.Config):
    project_root: Path | None = None
    """
    Root directory for this project.

    This isn't specific to the run; it is the parent directory of all runs.
    """

    logdir_basename: str = "nshtrainer"
    """Base name for the log directory."""

    setup_callback: DirectorySetupCallbackConfig = DirectorySetupCallbackConfig()
    """Configuration for the directory setup PyTorch Lightning callback."""

    def resolve_run_root_directory(self, run_id: str) -> Path:
        if (project_root_dir := self.project_root) is None:
            project_root_dir = Path.cwd()

        # The default base dir is $CWD/{logdir_basename}/{id}/
        base_dir = project_root_dir / self.logdir_basename
        base_dir.mkdir(exist_ok=True)

        # Add a .gitignore file to the {logdir_basename} directory
        #   which will ignore all files except for the .gitignore file itself
        gitignore_path = base_dir / ".gitignore"
        if not gitignore_path.exists():
            gitignore_path.touch()
            gitignore_path.write_text("*\n")

        base_dir = base_dir / run_id
        base_dir.mkdir(exist_ok=True)

        return base_dir

    def resolve_subdirectory(self, run_id: str, subdirectory: str) -> Path:
        # The subdir will be $CWD/{logdir_basename}/{id}/{log, stdio, checkpoint, activation}/
        if (subdir := getattr(self, subdirectory, None)) is not None:
            assert isinstance(subdir, Path), (
                f"Expected a Path for {subdirectory}, got {type(subdir)}"
            )
            return subdir

        dir = self.resolve_run_root_directory(run_id)
        dir = dir / subdirectory
        dir.mkdir(exist_ok=True)
        return dir

    def _resolve_log_directory_for_logger(self, run_id: str, logger: LoggerConfig):
        if (log_dir := logger.log_dir) is not None:
            return log_dir

        # Save to {logdir_basename}/{id}/log/{logger name}
        log_dir = self.resolve_subdirectory(run_id, "log")
        log_dir = log_dir / logger.resolve_logger_dirname()
        # ^ NOTE: Logger must have a `name` attribute, as this is
        # the discriminator for the logger registry
        log_dir.mkdir(exist_ok=True)

        return log_dir



from __future__ import annotations

import logging
from typing import Literal

from lightning.pytorch.utilities.exceptions import MisconfigurationException
from typing_extensions import final, override, assert_never
from lightning.pytorch import Trainer
from lightning.pytorch.callbacks import Callback
from ..metrics import MetricConfig
from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class MetricValidationCallbackConfig(CallbackConfigBase):
    name: Literal["metric_validation"] = "metric_validation"

    error_behavior: Literal["raise", "warn"] = "raise"
    """
    Behavior when an error occurs during validation:
    - "raise": Raise an error and stop the training.
    - "warn": Log a warning and continue the training.
    """

    validate_default_metric: bool = True
    """Whether to validate the default metric from the root config."""

    metrics: list[MetricConfig] = []
    """List of metrics to validate."""

    @override
    def create_callbacks(self, trainer_config):
        metrics = self.metrics.copy()
        if (
            self.validate_default_metric
            and (default_metric := trainer_config.primary_metric) is not None
        ):
            metrics.append(default_metric)

        yield MetricValidationCallback(self, metrics)


class MetricValidationCallback(Callback):
    def __init__(
        self,
        config: MetricValidationCallbackConfig,
        metrics: list[MetricConfig],
    ):
        super().__init__()

        self.config = config
        self.metrics = metrics

    def _check_metrics(self, trainer: Trainer):
        metric_names = ", ".join(metric.monitor for metric in self.metrics)
        log.info(f"Validating metrics: {metric_names}...")
        logged_metrics = set(trainer.logged_metrics.keys())

        invalid_metrics: list[str] = []
        for metric in self.metrics:
            if metric.monitor not in logged_metrics:
                invalid_metrics.append(metric.monitor)

        if invalid_metrics:
            msg = (
                f"The following metrics were not found in logged metrics: {invalid_metrics}\n"
                f"List of logged metrics: {list(trainer.logged_metrics.keys())}"
            )
            match self.config.error_behavior:
                case "raise":
                    raise MisconfigurationException(msg)
                case "warn":
                    log.warning(msg)
                case _:
                    assert_never(self.config.error_behavior)

    @override
    def on_sanity_check_end(self, trainer, pl_module):
        super().on_sanity_check_end(trainer, pl_module)

        self._check_metrics(trainer)

    @override
    def on_validation_end(self, trainer, pl_module):
        super().on_validation_end(trainer, pl_module)

        self._check_metrics(trainer)



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer.plugin.layer_sync import PluginConfigBase as PluginConfigBase
from nshtrainer.trainer.plugin.layer_sync import (
    TorchSyncBatchNormPlugin as TorchSyncBatchNormPlugin,
)
from nshtrainer.trainer.plugin.layer_sync import plugin_registry as plugin_registry

__all__ = [
    "PluginConfigBase",
    "TorchSyncBatchNormPlugin",
    "plugin_registry",
]
